#! /usr/bin/env python

"""
Module with various fits handling functions (same as in ``VIP``)
"""


__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['open_fits',
           'info_fits',
           'write_fits']


import os
import numpy as np
from astropy.io import fits as ap_fits


def open_fits(fitsfilename, n=0, header=False, ignore_missing_end=False,
              precision=np.float32, return_memmap=False, verbose=True, 
              **kwargs):
    """
    Load a fits file into a memory as numpy array.

    Parameters
    ----------
    fitsfilename : string or pathlib.Path
        Name of the fits file or ``pathlib.Path`` object
    n : int, optional
        It chooses which HDU to open. Default is the first one.
    header : bool, optional
        Whether to return the header along with the data or not.
    precision : numpy dtype, optional
        Float precision, by default np.float32 or single precision float.
    ignore_missing_end : bool optional
        Allows to open fits files with a header missing END card.
    return_memmap : bool, optional
        If True, the function returns the handle to the FITS file opened by
        mmap. With the hdulist, array data of each HDU to be accessed with mmap,
        rather than being read into memory all at once. This is particularly
        useful for working with very large arrays that cannot fit entirely into
        physical memory.
    verbose : bool, optional
        If True prints message of completion.
    **kwargs: optional
        Optional arguments to the astropy.io.fits.open() function. E.g. 
        "output_verify" can be set to ignore, in case of non-standard header.
        
    Returns
    -------
    hdulist : hdulist
        [memmap=True] FITS file ``n`` hdulist.
    data : numpy ndarray
        [memmap=False] Array containing the frames of the fits-cube.
    header : dict
        [memmap=False, header=True] Dictionary containing the fits header.

    """
    fitsfilename = str(fitsfilename)
    if not os.path.isfile(fitsfilename):
        fitsfilename += '.fits'

    hdulist = ap_fits.open(fitsfilename, ignore_missing_end=ignore_missing_end,
                           memmap=True, **kwargs)

    if return_memmap:
        return hdulist[n]
    else:
        data = hdulist[n].data
        data = np.array(data, dtype=precision)

        if header:
            header = hdulist[0].header
            if verbose:
                print("Fits HDU-{} data and header successfully loaded. "
                      "Data shape: {}".format(n, data.shape))
            return data, header
        else:
            if verbose:
                print("Fits HDU-{} data successfully loaded. "
                      "Data shape: {}".format(n, data.shape))
            return data


def info_fits(fitsfilename, **kwargs):
    """
    Print the information about a fits file.

    Parameters
    ----------
    fitsfilename : str
        Path to the fits file.
    **kwargs: optional
        Optional arguments to the astropy.io.fits.open() function. E.g. 
        "output_verify" can be set to ignore, in case of non-standard header.
        
    """
    with ap_fits.open(fitsfilename, memmap=True, **kwargs) as hdulist:
        hdulist.info()


def write_fits(fitsfilename, array, header=None, output_verify='exception',
               precision=np.float32, verbose=True):
    """
    Write array and header into FITS file.

    If there is a previous file with the same filename then it's replaced.

    Parameters
    ----------
    fitsfilename : string
        Full path of the fits file to be written.
    array : numpy ndarray
        Array to be written into a fits file.
    header : numpy ndarray, optional
        Array with header.
    output_verify : str, optional
        {"fix", "silentfix", "ignore", "warn", "exception"} 
        Verification options:
        https://docs.astropy.org/en/stable/io/fits/api/verification.html
    precision : numpy dtype, optional
        Float precision, by default np.float32 or single precision float.
    verbose : bool, optional
        If True prints message.

    """
    array = array.astype(precision, copy=False)
    if not fitsfilename.endswith('.fits'):
        fitsfilename += '.fits'

    if os.path.exists(fitsfilename):
        os.remove(fitsfilename)
        ap_fits.writeto(fitsfilename, array, header, output_verify)
        if verbose:
            print("Fits file successfully overwritten")
    else:
        ap_fits.writeto(fitsfilename, array, header, output_verify)
        if verbose:
            print("Fits file successfully saved")



#! /usr/bin/env python

"""
Function defining the goodness of fit.

.. [OLO16]
   | Olofsson et al. 2016
   | **Azimuthal asymmetries in the debris disk around HD 61005. A massive collision of planetesimals?**
   | *Astronomy & Astrophysics, Volume 591, p. 108*
   | `https://arxiv.org/abs/1601.07861
     <https://arxiv.org/abs/1601.07861>`_

.. [GRE16]
   | Greco & Brandt 2016
   | **The Measurement, Treatment, and Impact of Spectral Covariance and 
     Bayesian Priors in Integral-field Spectroscopy of Exoplanets**
   | *The Astrophysical Journal, Volume 833, Issue 1, p. 134*
   | `https://arxiv.org/abs/1602.00691
     <https://arxiv.org/abs/1602.00691>`_
     
"""

__author__ = 'Valentin Christiaens'
__all__ = ['goodness_of_fit',
           'gof_scal']

import numpy as np
from matplotlib import pyplot as plt
from .config import figsize
from .model_resampling import resample_model
from .utils_spec import extinction

def goodness_of_fit(lbda_obs, spec_obs, err_obs, lbda_mod, spec_mod, 
                    dlbda_obs=None, instru_corr=None, instru_res=None, 
                    instru_idx=None, use_weights=True, filter_reader=None, 
                    plot=False, outfile=None):
    """ Function to estimate the goodness of fit indicator defined as in 
    [OLO16]_ (Eq. 8). In addition, if a spectral correlation matrix is provided, 
    it is used to take into account the correlated noise between spectral 
    channels (see [GRE16]_). The goodness of fit indicator is identical to a 
    :math:`\chi^2` if all points are obtained with the same instrument (or 
    ``use_weights`` set to False) and in absence of spectrally correlated noise.

    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelengths of observed spectrum. If several instruments were used, the 
        wavelengths should be ordered per instrument, not necessarily as 
        monotonically increasing wavelength. Hereafter, :math:`n_{ch}` is the 
        length of ``lbda_obs``.
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of ``lbda_obs``. Should have a length
        of :math:`n_{ch}`.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. The array (list) can have either
        a length of :math:`n_{ch}`, or a shape of :math:`(2,n_{ch})` for lower 
        (first column) and upper (second column) uncertainties provided.
    lbda_mod : numpy 1d ndarray or list
        Wavelengths of tested model. Should have a wider (or equal) wavelength 
        extent than the observed spectrum.
    spec_mod : numpy 1d ndarray
        Model spectrum. It does not require the same wavelength sampling as 
        the observed spectrum. If higher spectral resolution, it will be 
        convolved with the instrumental spectral PSF (if ``instru_res`` is 
        provided) and then binned to the same sampling. If lower spectral 
        resolution, a linear interpolation is performed to infer the value at 
        the observed spectrum wavelength sampling.
    dlbda_obs : numpy 1d ndarray or list, optional
        Respective spectral channel width or FWHM of the photometric filters 
        used for each point of the observed spectrum. This vector is used to 
        infer which part(s) of a combined spectro+photometric spectrum should 
        involve convolution+subsampling (model resolution higher than 
        measurements), interpolation (the opposite), or convolution by the 
        transmission curve of a photometric filter. If not provided, it will be 
        inferred from the difference between consecutive lbda_obs points (i.e. 
        inaccurate for a combined spectrum). It must be provided IF one wants to 
        weigh each measurement based on the spectral resolution of each 
        instrument (as in [OLO16]_), through the ``use_weights`` argument.
    instru_corr : numpy 2d ndarray, optional
        Spectral correlation between post-processed images in which the 
        spectrum is measured. It is specific to the instrument, PSF subtraction 
        algorithm and radial separation of the companion from the central star.
        Can be computed using ``special.spec_corr.spectral_correlation``. In 
        case of a spectrum obtained with different instruments, it is 
        recommended to construct the final spectral_correlation matrix with
        ``special.spec_corr.combine_corrs``. If ``instru_corr`` is not provided, 
        the uncertainties in each spectral channel will be considered 
        independent. See [GRE16]_ for more details.
    instru_res : float or list of floats/strings, optional
        The mean instrumental resolving power(s) OR filter names. This is 
        used to convolve the model spectrum. If several instruments are used, 
        provide a list of resolving power values / filter names, one for 
        each instrument used.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to the number of instruments 
        (:math:`n_{ins}`). Zero for points that don't correspond to any of the 
        ``instru_res`` values provided, and i in :math:`[1,n_{ins}]` for points
        associated to instru_res[i-1]. This parameter must be provided if the 
        spectrum consists of points obtained with different instruments.
    use_weights: bool, optional
        For the likelihood calculation, whether to weigh each point of the 
        spectrum based on the spectral resolution or bandwidth of photometric
        filters used. Weights will be proportional to ``dlbda_obs/lbda_obs`` if 
        ``dlbda_obs`` is provided, or set to 1 for all points otherwise.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_res, the default file reader will be used. 
        It assumes the following format for the files:
            
        - first row contains headers (titles of each column)
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header \
        key name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer \
        or "lambda(nm)" for nanometer. Note: only what is in parentheses \
        matters for the units.
     
    plot : bool, optional
        Whether to plot the measured spectrum and the model spectrum.
    outfile : string, optional
        Path+filename for the plot to be saved (won't be saved if not provided).
        
    Returns
    -------
    chi_sq : float
        Goodness of fit indicator.

    """
    lbda_obs = np.array(lbda_obs)
    spec_obs = np.array(spec_obs)
    err_obs = np.array(err_obs)
    lbda_mod = np.array(lbda_mod)
    spec_mod = np.array(spec_mod)
    
    if lbda_obs.ndim != 1 or spec_obs.ndim != 1:
        raise TypeError('Observed lbda and spec must be lists or 1d arrays')
    if lbda_obs.shape[0] != spec_obs.shape[0]:
        raise TypeError('Obs lbda and spec need same shape')        
    if lbda_obs.shape[0] != err_obs.shape[-1]:
        raise TypeError('Obs lbda and err need same shape')
    if lbda_mod.ndim != 1 or spec_mod.ndim != 1:
        raise TypeError('The input model lbda/spec must be lists or 1d arrays')
    else:
        if lbda_mod.shape != spec_mod.shape:
            raise TypeError('The input model lbda and spec need same shape')
    if dlbda_obs is not None:
        if isinstance(dlbda_obs, list):
            dlbda_obs = np.array(dlbda_obs)
        if lbda_obs.shape != dlbda_obs.shape:
            raise TypeError('The input lbda_obs and dlbda_obs need same shape')

    n_ch = lbda_obs.shape[0]
             
    # interpolate OR convolve+bin model spectrum if not same sampling
    if len(lbda_obs) != len(lbda_mod):
        _,spec_mod_fin = resample_model(lbda_obs, lbda_mod, spec_mod, dlbda_obs, 
                                        instru_res, instru_idx, filter_reader)
    elif not np.allclose(lbda_obs, lbda_mod):
        _,spec_mod_fin = resample_model(lbda_obs, lbda_mod, spec_mod, dlbda_obs, 
                                        instru_res, instru_idx, filter_reader)
    else:
        spec_mod_fin = spec_mod
        
    # compute covariance matrix
    if err_obs.ndim == 2:
        err_obs = (np.absolute(err_obs[0])+np.absolute(err_obs[1]))/2.
    if instru_corr is None:
        instru_corr = np.diag(np.ones(n_ch))
    cov = np.ones_like(instru_corr)
    for ii in range(n_ch):
        for jj in range(n_ch):
            cov[ii,jj] = instru_corr[ii,jj]*err_obs[ii]*err_obs[jj] 


    # replace potential NaN values
    nnan_idx = np.where(np.isfinite(spec_mod_fin))
    if len(nnan_idx[0])>0:
        cov_crop = []
        for i in range(len(lbda_obs)):
            if i in nnan_idx[0]:
                tmp = cov[i]
                cov_crop.append(tmp[nnan_idx])
        cov = np.array(cov_crop)    
        spec_obs = spec_obs[nnan_idx]
        spec_mod_fin = spec_mod_fin[nnan_idx]
        if dlbda_obs is not None:
            dlbda_obs = dlbda_obs[nnan_idx]
        lbda_obs = lbda_obs[nnan_idx]
        
    delta = spec_obs-spec_mod_fin
    wi = np.ones_like(lbda_obs)
    
    if use_weights and dlbda_obs is not None:
        if np.sum(np.power((dlbda_obs[1:]/lbda_obs[1:])-(dlbda_obs[:-1]/lbda_obs[:-1]),2))!=0:
            # normalize weights for their sum to be equal to the number of points
            wi = np.sqrt(((dlbda_obs/lbda_obs)/np.sum(dlbda_obs/lbda_obs))*dlbda_obs.shape[0])    
   
    chi_sq = np.linalg.multi_dot((wi*delta,np.linalg.inv(cov),wi*delta))
                
    if plot:
        _, ax = plt.subplots(figsize=figsize)

        ax.plot(lbda_obs, lbda_obs*spec_obs, 'o', alpha=0.6, color='#1f77b4',
                label="Measured spectrum")
        ax.plot(lbda_obs, lbda_mod*spec_mod, '-', alpha=0.4, color='#1f77b4',
                label="Model")
        plt.xlabel('Wavelength')
        plt.ylabel(r'Flux density ($\lambda F_{\lambda}$)')

        plt.xlim(xmin=0.9*lbda_obs[0], xmax=1.1*lbda_obs[-1])
        plt.minorticks_on()
        plt.legend(fancybox=True, framealpha=0.5, fontsize=12, loc='best')
        plt.grid(which='major', alpha=0.2)
        if outfile:
            plt.savefig(outfile, bbox_inches='tight')
        plt.show()

    return chi_sq


def gof_scal(params, lbda_obs, spec_obs, err_obs, lbda_tmp, spec_tmp, dlbda_obs, 
             instru_corr, instru_res, instru_idx, use_weights, filter_reader, 
             ext_range):
    """ Wrapper of routine ``special.chi.goodness_of_fit`` for the goodness of 
    fit used to search for the best-fit library template spectrum to an observed
    spectrum. It has a ``params`` argument to consider the scaling factor and 
    optionnally the differential extinction as free parameter(s).
    
    Parameters
    ----------
    params: tuple
        Tuple of 1 or 2 elements: scaling factor and (optionally) differential
        optical extinction :math:`\Delta A_V` (:math:`\Delta A_V` can be 
        negative if template spectra are not dereddened).
    lbda_obs : numpy 1d ndarray or list
        Wavelengths of observed spectrum. If several instruments were used, the 
        wavelengths should be ordered per instrument, not necessarily as 
        monotonically increasing wavelength. Hereafter, :math:`n_{ch}` is the 
        length of ``lbda_obs``.
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of ``lbda_obs``. Should have a length
        of :math:`n_{ch}`.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. The array (list) can have either a 
        length of :math:`n_{ch}`, or a shape of :math:`(2,n_{ch})` for lower 
        (first column) and upper (second column) uncertainties provided.
    lbda_tmp : numpy 1d ndarray or list
        Wavelengths of tested template. Should have a wider wavelength extent 
        than the observed spectrum.
    spec_tmp : numpy 1d ndarray
        Template spectrum. It does not require the same wavelength sampling as 
        the observed spectrum. If higher spectral resolution, it will be 
        convolved with the instrumental spectral PSF (if ``instru_res`` is 
        provided) and then binned to the same sampling. If lower spectral 
        resolution, a linear interpolation is performed to infer the value at 
        the observed spectrum wavelength sampling.
    dlbda_obs : numpy 1d ndarray or list, optional
        Respective spectral channel width or FWHM of the photometric filters 
        used for each point of the observed spectrum. This vector is used to 
        infer which part(s) of a combined spectro+photometric spectrum should 
        involve convolution+subsampling (model resolution higher than 
        measurements), interpolation (the opposite), or convolution by the 
        transmission curve of a photometric filter. If not provided, it will be 
        inferred from the difference between consecutive lbda_obs points (i.e. 
        inaccurate for a combined spectrum). It must be provided IF one wants to 
        weigh each measurement based on the spectral resolution of each 
        instrument (as in [OLO16]_), through the ``use_weights`` argument.
    instru_corr : numpy 2d ndarray, optional
        Spectral correlation between post-processed images in which the 
        spectrum is measured. It is specific to the instrument, PSF subtraction 
        algorithm and radial separation of the companion from the central star.
        Can be computed using ``special.spec_corr.spectral_correlation``. In 
        case of a spectrum obtained with different instruments, it is 
        recommended to construct the final spectral_correlation matrix with
        ``special.spec_corr.combine_corrs``. If ``instru_corr`` is not provided, 
        the uncertainties in each spectral channel will be considered 
        independent. See [GRE16]_ for more details.
    instru_res : float or list of floats/strings, optional
        The mean instrumental resolving power(s) OR filter names. This is 
        used to convolve the model spectrum. If several instruments are used, 
        provide a list of resolving power values / filter names, one for 
        each instrument used.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to the number of instruments 
        (:math:`n_{ins}`). Zero for points that don't correspond to any of the 
        ``instru_res`` values provided, and i in :math:`[1,n_{ins}]` for points
        associated to instru_res[i-1]. This parameter must be provided if the 
        spectrum consists of points obtained with different instruments.
    use_weights: bool, optional
        For the likelihood calculation, whether to weigh each point of the 
        spectrum based on the spectral resolution or bandwidth of photometric
        filters used. Weights will be proportional to ``dlbda_obs/lbda_obs`` if 
        ``dlbda_obs`` is provided, or set to 1 for all points otherwise.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_res, the default file reader will be used. 
        It assumes the following format for the files:
            
        - first row contains headers (titles of each column)
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header \
        key name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer \
        or "lambda(nm)" for nanometer. Note: only what is in parentheses \
        matters for the units.
             
    ext_range : tuple or None, opt
        - If None: differential extinction is not considered as a free parameter. 
        - If a tuple: it should contain 2 floats (for simplex \
        ``search_mode``) or 3 floats (for grid search ``search_mode``) \
        corresponding to the lower limit, upper limit (and step for the grid \
        search). For the simplex search, the lower and upper limits are used \
        to set a chi square of infinity outside of the range.
        
    Returns
    -------
    chi_sq : float
        Goodness of fit indicator.     
    
    Note
    ----
    If several filter filenames are provided in ``instru_res``, the filter files
    must all have the same format and wavelength units (for reading by the same
    ``filter_reader`` snippet or default function).
        
    """
    tmp_spec = spec_tmp*params[0]
    
    if len(params) == 2:
        if ext_range is None:
            raise TypeError("ext_range should be a tuple of 3 elements")
        else:
            if params[1]<ext_range[0] or params[1]>ext_range[1]:
                return np.inf
            AV=params[1]
            if len(ext_range) == 3 and abs(AV) < ext_range[-1]: # round to zero if smaller than step
                AV = 0
            Albdas = extinction(lbda_tmp,abs(AV))
            extinc_fac = np.power(10.,-Albdas/2.5)
            if AV>0:
                tmp_spec *= extinc_fac
            elif AV<0:
                tmp_spec /= extinc_fac
    elif len(params) > 2:
        raise TypeError("params tuple should have length 1 or 2")
    
    
    return goodness_of_fit(lbda_obs, spec_obs, err_obs, lbda_tmp, tmp_spec, 
                           dlbda_obs=dlbda_obs, instru_corr=instru_corr, 
                           instru_res=instru_res, instru_idx=instru_idx, 
                           use_weights=use_weights, filter_reader=filter_reader)


#! /usr/bin/env python

"""
Module with functions for posterior sampling of the model spectra parameters
using nested sampling (either ``nestle`` or ``ultranest`` samplers).

.. [nes13]
   | Barbary 2013
   | **nestle**
   | *GitHub*
   | `https://github.com/kbarbary/nestle
     <https://github.com/kbarbary/nestle>`_

.. [SKI04]
   | Skilling 2004
   | **Bayesian Inference and Maximum Entropy Methods in Science and Engineering:
     24th International Workshop on Bayesian Inference and Maximum Entropy
     Methods in Science and Engineering**
   | *American Institute of Physics Conference Series, Volume 735, pp. 395-405*
   | `https://ui.adsabs.harvard.edu/abs/2004AIPC..735..395S
     <https://ui.adsabs.harvard.edu/abs/2004AIPC..735..395S>`_

.. [MUK06]
   | Mukherjee et al. 2006
   | **A Nested Sampling Algorithm for Cosmological Model Selection**
   | *ApJL, Volume 638, Issue 2, pp. 51-54*
   | `https://arxiv.org/abs/astro-ph/0508461
     <https://arxiv.org/abs/astro-ph/0508461>`_

.. [FER09]
   | Feroz et al. 2009
   | **MULTINEST: an efficient and robust Bayesian inference tool for cosmology
     and particle physics**
   | *MNRAS, Volume 398, Issue 4, pp. 1601-1614*
   | `https://arxiv.org/abs/0809.3437
     <https://arxiv.org/abs/0809.3437>`_

.. [BUC21]
   | Buchner 2021
   | **UltraNest - a robust, general purpose Bayesian inference engine**
   | *JOSS, Volume 6, Issue 60, p. 3001*
   | `https://arxiv.org/abs/2101.09604
     <https://arxiv.org/abs/2101.09604>`_

"""



__author__ = 'V. Christiaens',
__all__ = ['nested_spec_sampling',
           'show_nestle_results',
           'show_ultranest_results']

from astropy import constants as con
import nestle
import ultranest
from ultranest.plot import cornerplot, PredictionBand
import numpy as np
from matplotlib import pyplot as plt
import pandas as pd

from .config import time_ini, timing
from .fits import open_fits, write_fits
from .mcmc_sampling import lnlike, confidence, show_walk_plot, show_corner_plot
from .model_resampling import make_resampled_models, make_model_from_params
from .utils_nested import un_burning
from os.path import isfile
from scipy.special import ndtri


def nested_spec_sampling(init, lbda_obs, spec_obs, err_obs, dist,
                         grid_param_list, labels, bounds, resamp_before=True,
                         model_grid=None, model_reader=None, em_lines={},
                         em_grid={}, dlbda_obs=None, instru_corr=None,
                         instru_res=None, instru_idx=None, use_weights=True,
                         filter_reader=None, AV_bef_bb=False, units_obs='si',
                         units_mod='si', interp_order=1, priors=None,
                         physical=True, interp_nonexist=True,
                         output_dir='special/', grid_name='resamp_grid.fits',
                         sampler='ultranest', method='single', npoints=100,
                         dlogz=0.2, verbose=True, **kwargs):


    """ Runs a nested sampling algorithm in order to retrieve the best-fit
    parameters for given spectral model and observed spectrum.. The
    result of this procedure is either a ``nestle`` [nes13]_ or ``ultranest``
    [BUC21]_ object (depending on the chosen sampler) containing the samples
    from the posterior distributions of each of the parameters.
    For the ``nestle`` sampler, several methods are available corresponding to
    MCMC [SKI04]_, single ellipsoid [MUK06]_ or multiple ellipsoid [FER09]_.

    Parameters
    ----------
    init: numpy ndarray or tuple
        Initial guess on the best fit parameters of the spectral fit. Length of
        the tuple should match the total number of free parameters.
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the intensity of emission lines (labels must match \
        those in the em_lines dict), in units of the model spectrum (x mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb \
        contribution
    lbda_obs : numpy 1d ndarray or list
        Wavelengths of observed spectrum. If several instruments were used, the
        wavelengths should be ordered per instrument, not necessarily as
        monotonically increasing wavelength. Hereafter, :math:`n_{ch}` is the
        length of ``lbda_obs``.
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of ``lbda_obs``. Should have a length
        of :math:`n_{ch}`.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. The array (list) can have either
        a length of :math:`n_{ch}`, or a shape of :math:`(2,n_{ch})` for lower
        (first column) and upper (second column) uncertainties provided.
    dist :  float
        Distance in parsec, used for flux scaling of the models.
    grid_param_list : list of 1d numpy arrays/lists
        Should contain list/numpy 1d arrays with available grid of model
        parameters (should only contain the sampled parameters, not the models
        themselves). The models will be loaded with ``model_reader``.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state:
            - first all parameters related to loaded models (e.g. 'Teff', 'logg')
            - then the planet photometric radius 'R', in Jupiter radius
            - (optionally) the flux of emission lines (labels should match those \
            in the em_lines dictionary), in units of the model spectrum (times mu)
            - (optionally) the optical extinction 'Av', in mag
            - (optionally) the ratio of total to selective optical extinction 'Rv'
            - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb \
            contribution.
    bounds: dictionary
        Each entry should be associated with a tuple corresponding to lower and
        upper bounds respectively. Bounds should be provided for ALL model
        parameters, including 'R' (planet photometric radius). 'Av' (optical
        extinction) is optional. If provided here, Av will also be fitted.
        All keywords that are neither 'R', 'Av' nor 'M' will
        be considered model grid parameters.

        Examples:
            >>> bounds = {'Teff':(1000,2000), 'logg':(3.0,4.5), 'R':(0.1,5)}
            >>> bounds = {'Teff':(1000,2000), 'logg':(3.0,4.5), 'R':(0.1,5),
            >>>           'Av':(0.,2.5), 'Rv':(1,5)}
            >>> bounds = {'Teff':(1000,2000), 'logg':(3.0,4.5), 'R':(0.1,5),
            >>>           'Av':(0.,2.5), 'BrG':(1e-17,1e-15)}
            >>> bounds = {'Teff':(2000,3000), 'logg':(3.0,4.5), 'R':(1.,5.),
            >>>           'Tbb1':(500,1500), 'Rbb1':(0.1,1.)}

    resamp_before: bool, optional
        Whether to prepare the whole grid of resampled models before entering
        the MCMC, i.e. to avoid doing it at every MCMC step. Recommended.
        Only reason not to: model grid is too large and individual models
        require being opened and resampled at each step.
    model_grid : numpy N-d array, optional
        If provided, should contain the grid of model spectra for each
        free parameter of the given grid. I.e. for a grid of :math:`n_T` values
        of :math:`T_{eff}` and :math:`n_g` values of log(:math:`g`), the numpy
        array should have a shape of :math:`(n_T, n_g, n_{ch}, 2)`, where the
        last 2 dimensions correspond to wavelength and fluxes respectively. If
        provided, ``model_grid`` takes precedence over ``model_name``/
        ``model_reader``.
    model_reader : python routine, opt
        External routine that reads a model file and returns a 2D numpy array,
        where the first column corresponds to wavelengths, and the second
        contains model values. See example routine in
        ``special.model_resampling.interpolate_model`` description.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        4 values:

            1. the wavelength (in :math:`\mu` m);

            2. a string indicating whether line intensity is expressed in flux \
            ('F'), luminosity ('L') or log(L/LSun) ("LogL");

            3. the FWHM of the gaussian (or None if to be set automatically);

            4. whether the FWHM is expressed in 'nm', 'mu' or 'km/s'.

        The third and fourth can also be set to None. In that case, the FWHM of
        the gaussian will automatically be set to the equivalent width of the
        line, calculated from the flux to be injected and the continuum
        level (measured in the grid model to which the line is injected).

        Examples:
            >>> em_lines = {'BrG':(2.1667,'F',None, None)};
            >>> em_lines = {'BrG':(2.1667,'LogL', 100, 'km/s')}

    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation
        efficiency, interpolation will be performed between the points of this
        grid during the MCMC sampling. Dictionary entries should match those in
        ``labels`` and ``em_lines``.

        Examples:
            >>> BrGmin, BrGmax = -5, 5
            >>> em_grid = {'BrG': np.arange(BrGmin, BrGmax, 20)}

            >>> BrGmin, BrGmax = -5, 5
            >>> PaBmin, PaBmax = -2, 7
            >>> em_grid = {'PaB': np.arange(PaBmin, PaBmax, 20),
            >>>            'BrG': np.arange(BrGmin, BrGmax, 20)}

    dlbda_obs : numpy 1d ndarray or list, optional
        Respective spectral channel width or FWHM of the photometric filters
        used for each point of the observed spectrum. This vector is used to
        infer which part(s) of a combined spectro+photometric spectrum should
        involve convolution+subsampling (model resolution higher than
        measurements), interpolation (the opposite), or convolution by the
        transmission curve of a photometric filter. If not provided, it will be
        inferred from the difference between consecutive lbda_obs points (i.e.
        inaccurate for a combined spectrum). It must be provided IF one wants to
        weigh each measurement based on the spectral resolution of each
        instrument (as in [OLO16]_), through the ``use_weights`` argument.
    instru_corr : numpy 2d ndarray, optional
        Spectral correlation between post-processed images in which the
        spectrum is measured. It is specific to the instrument, PSF subtraction
        algorithm and radial separation of the companion from the central star.
        Can be computed using ``special.spec_corr.spectral_correlation``. In
        case of a spectrum obtained with different instruments, it is
        recommended to construct the final spectral_correlation matrix with
        ``special.spec_corr.combine_corrs``. If ``instru_corr`` is not provided,
        the uncertainties in each spectral channel will be considered
        independent. See [GRE16]_ for more details.
    instru_res : float or list of floats/strings, optional
        The mean instrumental resolving power(s) OR filter names. This is
        used to convolve the model spectrum. If several instruments are used,
        provide a list of resolving power values / filter names, one for
        each instrument used.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used
        to obtain the spectrum, label them from 0 to the number of instruments
        (:math:`n_{ins}`). Zero for points that don't correspond to any of the
        ``instru_res`` values provided, and i in :math:`[1,n_{ins}]` for points
        associated to instru_res[i-1]. This parameter must be provided if the
        spectrum consists of points obtained with different instruments.
    use_weights: bool, optional
        For the likelihood calculation, whether to weigh each point of the
        spectrum based on the spectral resolution or bandwidth of photometric
        filters used. Weights will be proportional to ``dlbda_obs/lbda_obs`` if
        ``dlbda_obs`` is provided, or set to 1 for all points otherwise.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array,
        where the first column corresponds to wavelengths, and the second
        contains transmission values. Important: if not provided, but strings
        are detected in instru_res, the default file reader will be used.
        It assumes the following format for the files:

        - first row contains headers (titles of each column)
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header \
        key name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer \
        or "lambda(nm)" for nanometer. Note: only what is in parentheses \
        matters for the units.

    AV_bef_bb: bool, optional
        If both extinction and an extra bb component are free parameters,
        whether to apply extinction before adding the BB component (e.g.
        extinction mostly from circumplanetary dust) or after the BB component
        (e.g. mostly insterstellar extinction).
    units_obs : str, opt {'si','cgs','jy'}
        Units of observed spectrum. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu
        or 'jy' for janskys.
    units_mod: str, opt {'si','cgs','jy'}
        Units of the model. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu or 'jy'
        for janskys. If different to units_obs, the spectrum units will be
        converted.
    interp_order: int or tuple of int, optional, {-1,0,1}
        Interpolation mode for model interpolation. If a tuple of integers, the
        length should match the number of grid dimensions and will trigger a
        different interpolation mode for the different parameters.
            - -1: Order 1 spline interpolation in logspace for the parameter
            - 0: nearest neighbour model
            - 1: Order 1 spline interpolation

    priors: dictionary, opt
        If not None, sets prior estimates for each parameter of the model. Each
        entry should be set to either None (no prior) or a tuple of 2 elements
        containing prior estimate and uncertainty on the estimate.
        Missing entries (i.e. provided in ``bounds`` dictionary but not here)
        will be associated no prior.
        'M' can be used for a prior on the mass of the planet. In that case the
        corresponding prior log probability is computed from the values for
        parameters 'logg' and 'R'.

        Examples:
            >>> priors = {'logg':(3.5,0.2), 'Rv':(3.1,0.2)}
            >>> priors = {'Teff':(1600,100), 'logg':(3.5,0.5), 'R':(1.6,0.1),
            >>>           'Av':(1.8,0.2), 'M':(10,3)}

        Important: dictionary entry names should match exactly those of
        ``bounds``.
    physical: bool, opt
        In case of extra black body component(s) to a photosphere, whether to
        force lower temperature than the photosphere effective temperature.
    interp_nonexist: bool, opt
        Whether to interpolate non-existing models in the grid. Only used if
        resamp_before is set to True.
    w : float or tuple
        The relative size of the bounds (around the initial state ``init``) for
        each parameter. If a float the same relative size is considered for
        each parameter. E.g. if 0.1, bounds will be set to:
        (0.9*params[0], 1.1*params[0]),
        ...
        (0.9*params[N-1], 1.1*params[N-1]),
        to True), or make it and write it if it does not.
    output_dir: str, optional
        The name of the output directory which contains the output files in the
        case  ``save`` is True.
    grid_name: str, optional
        Name of the fits file containing the model grid (numpy array) AFTER
        convolution+resampling as the observed spectrum given as input.
        If provided, will read it if it exists (and resamp_before is set
    method : {"single", "multi", "classic"}, str optional
        Flavor of nested sampling.
    npoints : int optional
        Number of active points. Must be >> ndim+1, otherwise will produce bad
        results. For UltraNest, this is the minimum number of active points.
    dlogz : float, optional
        Target evidence uncertainty. Iterations will stop when the estimated
        contribution of the remaining prior volume to the total evidence falls
        below this threshold. Explicitly, the stopping criterion is
        log(z + z_est) - log(z) < dlogz
        where z is the current evidence from all saved samples, and z_est is the
        estimated contribution from the remaining volume. This option and
        decline_factor are mutually exclusive. If neither is specified, the
        default is dlogz=0.2.
    decline_factor : float, optional
        If supplied, iteration will stop when the weight (likelihood times prior
        volume) of newly saved samples has been declining for
        decline_factor * nsamples consecutive samples. A value of 1.0 seems to
        work pretty well.
    rstate : random instance, optional
        RandomState instance. If not given, the global random state of the
        numpy.random module will be used.
    kwargs: optional
        Additional optional arguments to either the `nestle.sample` or
        `ultranest.ReactiveNestedSampler.run` functions.

    Returns
    -------
    res : nestle object
        ``Nestle`` object with the nested sampling results, including the
        posterior samples.

    Notes
    -----
    Nested Sampling is a computational approach for integrating posterior
    probability in order to compare models in Bayesian statistics. It is similar
    to Markov Chain Monte Carlo (MCMC) in that it generates samples that can be
    used to estimate the posterior probability distribution. Unlike MCMC, the
    nature of the sampling also allows one to calculate the integral of the
    distribution. It also happens to be a pretty good method for robustly
    finding global maxima.

    Nestle documentation:
        http://kbarbary.github.io/nestle/

    Convergence:
        http://kbarbary.github.io/nestle/stopping.html
        Nested sampling has no well-defined stopping point. As iterations continue,
        the active points sample a smaller and smaller region of prior space.
        This can continue indefinitely. Unlike typical MCMC methods, we don't gain
        any additional precision on the results by letting the algorithm run longer;
        the precision is determined at the outset by the number of active points.
        So, we want to stop iterations as soon as we think the active points are
        doing a pretty good job sampling the remaining prior volume - once we've
        converged to the highest-likelihood regions such that the likelihood is
        relatively flat within the remaining prior volume.

    Method:
        The trick in nested sampling is to, at each step in the algorithm,
        efficiently choose a new point in parameter space drawn with uniform
        probability from the parameter space with likelihood greater than the
        current likelihood constraint. The different methods all use the
        current set of active points as an indicator of where the target
        parameter space lies, but differ in how they select new points from it:

        - "classic" is close to the method described in [SKI04]_.
        - "single" [MUK06]_ determines a single ellipsoid that bounds all active \
        points, enlarges the ellipsoid by a user-settable factor, and \
        selects a new point at random from within the ellipsoid.
        - "multiple" [FER09]_ (Multinest). In cases where the posterior is \
        multi-modal, the single-ellipsoid method can be extremely inefficient. \
        In such situations, there are clusters of active points on separate \
        high-likelihood regions separated by regions of lower likelihood. Bounding \
        all points in a single ellipsoid means that the ellipsoid includes the \
        lower-likelihood regions we wish to avoid sampling from.\
        The solution is to detect these clusters and bound them in separate\
        ellipsoids. For this, we use a recursive process where we perform\
        K-means clustering with K=2. If the resulting two ellipsoids have a\
        significantly lower total volume than the parent ellipsoid (less than half),\
        we accept the split and repeat the clustering and volume test on each of\
        the two subset of points. This process continues recursively.\
        Alternatively, if the total ellipse volume is significantly greater\
        than expected (based on the expected density of points) this indicates\
        that there may be more than two clusters and that K=2 was not an\
        appropriate cluster division.\
        We therefore still try to subdivide the clusters recursively. However,\
        we still only accept the final split into N clusters if the total volume\
        decrease is significant.

    Note
    ----
    - If several filter filenames are provided in ``instru_res``, the filter files
      must all have the same format and wavelength units (for reading by the same
      ``filter_reader`` snippet or default function).

    - ``grid_param_list`` and ``model_grid`` shouldn't contain grids on radius
      and Av. For a combined grid model + black body fit, just provide the list
      of parameters probed by the grid to ``grid_param_list``, and provide values
      for 'Tbbn' and 'Rbbn' to ``initial_state``, ``labels`` and ``bounds``.

    """
    # ----------------------- Preparation/Formatting --------------------------

    nparams = len(init)

    if grid_param_list is not None:
        if model_grid is None and model_reader is None:
            msg = "Either model_grid or model_reader have to be provided"
            raise TypeError(msg)
        n_gparams = len(grid_param_list)
        gp_dims = []
        for nn in range(n_gparams):
            gp_dims.append(len(grid_param_list[nn]))
        gp_dims = tuple(gp_dims)
    else:
        n_gparams = 0

    # format emission line dictionary and em grid
    if len(em_grid)>0:
        n_em = len(em_grid)
        em_dims = []
        for lab in labels:
            if lab in em_grid.keys():
                em_dims.append(len(em_grid[lab]))
        em_dims = tuple(em_dims)
        # update the grids depending on input units => make it surface flux
        idx_R = labels.index('R')
        for key, val in em_lines.items():
            if val[1] == 'L':
                idx_line = labels.index(key)
                # adapt grid
                R_si = init[idx_R]*con.R_jup.value
                conv_fac = 4*np.pi*R_si**2
                tmp = np.array(em_grid[key])/conv_fac
                em_grid[key] = tmp.tolist()
                # adapt ini state
                init = list(init)
                init[idx_line] /= conv_fac
                init = tuple(init)
                #adapt bounds
                bounds_ori = list(bounds[key])
                bounds[key] = (bounds_ori[0]/conv_fac, bounds_ori[1]/conv_fac)
            elif val[1] == 'LogL':
                idx_line = labels.index(key)
                R_si = init[idx_R]*con.R_jup.value
                conv_fac = con.L_sun.value/(4*np.pi*R_si**2)
                tmp = np.power(10,np.array(em_grid[key]))*conv_fac
                em_grid[key] = tmp.tolist()
                # adapt ini state
                init = list(init)
                init[idx_line] = conv_fac*10**init[idx_line]
                init = tuple(init)
                #adapt bounds
                bounds_ori = list(bounds[key])
                bounds[key] = (conv_fac*10**bounds_ori[0],
                               conv_fac*10**bounds_ori[1])
            if em_lines[key][2] is not None:
                if em_lines[key][-1] == 'km/s':
                    v = em_lines[key][2]
                    em_lines_tmp = list(em_lines[key])
                    em_lines_tmp[2] = (1000*v/con.c.value)*em_lines[key][0]
                    em_lines_tmp[3] = 'mu'
                    em_lines[key] = tuple(em_lines_tmp)
                elif em_lines[key][-1] == 'nm':
                    em_lines_tmp = list(em_lines[key])
                    em_lines_tmp[2] = em_lines[key][2]/1000
                    em_lines_tmp[3] = 'mu'
                    em_lines[key] = tuple(em_lines_tmp)
                elif em_lines[key][-1] != 'mu':
                    msg = "Invalid unit of FWHM for line injection"
                    raise ValueError(msg)

    if model_grid is not None and grid_param_list is not None:
        if model_grid.ndim-2 != n_gparams:
            msg = "Ndim-2 of model_grid should match len(grid_param_list)"
            raise TypeError(msg)


    # Check model grid parameters extend beyond bounds to avoid extrapolation
    if grid_param_list is not None:
        for pp in range(n_gparams):
            if grid_param_list[pp][0]>bounds[labels[pp]][0]:
                msg= "Grid has to extend beyond bounds for {}."
                msg+="\n Consider increasing the lower bound to >{}."
                raise ValueError(msg.format(labels[pp],grid_param_list[pp][0]))
            if grid_param_list[pp][-1]<bounds[labels[pp]][1]:
                msg= "Grid has to extend beyond bounds for {}."
                msg+="\n Consider decreasing the upper bound to <{}."
                raise ValueError(msg.format(labels[pp],grid_param_list[pp][1]))

    # Check initial state is within bounds for all params (not only grid)
    for pp in range(nparams):
        if init[pp]<bounds[labels[pp]][0]:
            msg= "Initial state has to be within bounds for {}."
            msg+="\n Consider decreasing the lower bound to <{}."
            raise ValueError(msg.format(labels[pp],init[pp]))
        if init[pp]>bounds[labels[pp]][1]:
            msg= "Initial state has to be within bounds for {}."
            msg+="\n Consider decreasing the upper bound to >{}."
            raise ValueError(msg.format(labels[pp],init[pp]))

    # Prepare model grid: convolve+resample models as observations
    if resamp_before and grid_param_list is not None:
        if isfile(output_dir+grid_name):
            model_grid = open_fits(output_dir+grid_name)
            # check its shape is consistent with grid_param_list
            if model_grid.shape[:n_gparams] != gp_dims:
                msg="the loaded model grid ({}) doesn't have expected dims ({})"
                raise TypeError(msg.format(model_grid.shape,gp_dims))
            elif model_grid.shape[-2] != len(lbda_obs):
                msg="the loaded model grid doesn't have expected WL dimension"
                raise TypeError(msg)
            elif model_grid.shape[-1] != 2:
                msg="the loaded model grid doesn't have expected last dimension"
                raise TypeError(msg)
            elif len(em_grid) > 0:
                if model_grid.shape[n_gparams:n_gparams+n_em] != em_dims:
                    msg="loaded model grid ({}) doesn't have expected dims ({})"
                    raise TypeError(msg.format(model_grid.shape,em_dims))
        else:
            model_grid = make_resampled_models(lbda_obs, grid_param_list,
                                               model_grid, model_reader,
                                               em_lines, em_grid, dlbda_obs,
                                               instru_res, instru_idx,
                                               filter_reader, interp_nonexist)
            if output_dir and grid_name:
                write_fits(output_dir+grid_name, model_grid)
        # note: if model_grid is provided, it is still resampled to the
        # same wavelengths as observed spectrum. However, if a fits name is
        # provided in grid_name and that file exists, it is assumed the model
        # grid in it is already resampled to match lbda_obs.


    # -------------- Definition of utilities for nested sampling --------------

    def prior_transform(x):
        """
        Computes the transformation from the unit distribution `[0, 1]` to
        parameter space. Uniform distributions are assumed for all parameters.


        Notes
        -----
        The prior transform function is used to specify the Bayesian prior for
        the problem, in a round-about way. It is a transformation from a space
        where variables are independently and uniformly distributed between 0
        and 1 to the parameter space of interest. For independent parameters,
        this would be the product of the inverse cumulative distribution
        function (also known as the percent point function or quantile function)
        for each parameter.
        http://kbarbary.github.io/nestle/prior.html
        """
        # uniform priors
        pt = []
        for p in range(nparams):
            pmin = bounds[labels[p]][0]
            pmax = bounds[labels[p]][1]
            pt.append(x[p] * (pmax - pmin) + pmin)

        if priors is not None:
            # replace with Gaussian prior where relevant
            for key, prior in priors.items():
                if key == 'M' and 'logg' in labels:
                    msg = "Mass prior only available for MCMC sampling"
                    raise ValueError(msg)
                else:
                    idx_prior = labels.index(key)
                    pt[idx_prior] = prior[0] + prior[1] * ndtri(x)

        return np.array(pt)

    def f(param):

        return lnlike(param, labels, grid_param_list, lbda_obs, spec_obs,
                      err_obs, dist, model_grid=model_grid,
                      model_reader=model_reader, em_lines=em_lines,
                      em_grid=em_grid, dlbda_obs=dlbda_obs,
                      instru_corr=instru_corr, instru_res=instru_res,
                      instru_idx=instru_idx, use_weights=use_weights,
                      filter_reader=filter_reader, AV_bef_bb=AV_bef_bb,
                      units_obs=units_obs, units_mod=units_mod,
                      interp_order=interp_order)


    # ------------------------ Actual sampling --------------------------------
    if verbose:  start = time_ini()

    if verbose:
        print('Prior bounds on parameters:')
        for p in range(nparams):
            pmin = bounds[labels[p]][0]
            pmax = bounds[labels[p]][1]
            print('{} [{},{}]'.format(labels[p], pmin, pmax))
        print('\nUsing {} active points'.format(npoints))


    if sampler == 'nestle':
        final_res = nestle.sample(f, prior_transform, ndim=nparams,
                                  method=method, npoints=npoints, dlogz=dlogz,
                                  **kwargs)
    elif sampler == 'ultranest':
        if isinstance(labels, tuple):
            labels = list(labels)
        un_sampler = ultranest.ReactiveNestedSampler(labels, f, prior_transform,
                                                     log_dir=output_dir,
                                                     resume=True)
        res = un_sampler.run(min_num_live_points=npoints, dlogz=dlogz, **kwargs)
        final_res = (un_sampler, res)

    if verbose:
        print('\nTotal running time:')
        timing(start)

    return final_res


def show_nestle_results(ns_object, labels, method, burnin=0.4, bins=None,
                        cfd=68.27, units=None, ndig=None, labels_plot=None,
                        save=False, output_dir='/', plot=False,  **kwargs):
    """
    Show the results obtained with the Nestle sampler: summary, parameters with
    errors, walk and corner plots. Returns best-fit values and uncertatinties.

    Parameters
    ----------
    ns_object: numpy.array
        The nestle object returned from `nested_spec_sampling`.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state:
            - first all parameters related to loaded models (e.g. 'Teff', 'logg')
            - then the planet photometric radius 'R', in Jupiter radius
            - (optionally) the flux of emission lines (labels should match those \
            in the em_lines dictionary), in units of the model spectrum (times mu)
            - (optionally) the optical extinction 'Av', in mag
            - (optionally) the ratio of total to selective optical extinction 'Rv'
            - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb \
            contribution.
    method : {"single", "multi", "classic"}, str optional
        Flavor of nested sampling.
    burnin: float, default: 0
        The fraction of a walker we want to discard.
    bins: int, optional
        The number of bins used to sample the posterior distributions.
    cfd: float, optional
        The confidence level given in percentage.
    units: tuple, opt
        Tuple of strings containing units for each parameter. If provided,
        mcmc_res will be printed on top of each 1d posterior distribution along
        with these units.
    ndig: tuple, opt
        Number of digits precision for each printed parameter.
    labels_plot: tuple, opt
        Labels corresponding to parameter names, used for the plot. If None,
        will use "labels" passed in kwargs.
    save: boolean, default: False
        If True, a pdf file is created.
    output_dir: str, optional
        The name of the output directory which contains the output files in the
        case  ``save`` is True.
    plot: bool, optional
        Whether to show the plots (instead of saving them).
    kwargs:
        Additional optional arguments passed to `confidence` (matplotlib
        optional arguments for histograms).

    Returns
    -------
    final_res: numpy ndarray
         Best-fit parameters and uncertainties (corresponding to 68% confidence
         interval). Dimensions: nparams x 2.

    """
    res = ns_object
    nsamples = res.samples.shape[0]
    indburnin = int(np.percentile(np.array(range(nsamples)), burnin * 100))
    nparams = len(labels)

    print(res.summary())

    print(
        '\nNatural log of prior volume and Weight corresponding to each sample')
    if save or plot:
        plt.figure(figsize=(12, 4))
        plt.subplot(1, 2, 1)
        plt.plot(res.logvol, '.', alpha=0.5, color='gray')
        plt.xlabel('samples')
        plt.ylabel('logvol')
        plt.vlines(indburnin, res.logvol.min(), res.logvol.max(),
                   linestyles='dotted')
        plt.subplot(1, 2, 2)
        plt.plot(res.weights, '.', alpha=0.5, color='gray')
        plt.xlabel('samples')
        plt.ylabel('weights')
        plt.vlines(indburnin, res.weights.min(), res.weights.max(),
                   linestyles='dotted')
        if save:
            plt.savefig(output_dir+'Nestle-{}_results.pdf'.format(method))
        if plot:
            plt.show()

        print("\nWalk plots before the burnin")
        show_walk_plot(np.expand_dims(res.samples, axis=0), labels)
        if burnin > 0:
            print("\nWalk plots after the burnin")
            show_walk_plot(np.expand_dims(res.samples[indburnin:], axis=0),
                           labels)
        if save:
            plt.savefig(output_dir+'Nestle-{}_walk_plots.pdf'.format(method))

    mean, cov = nestle.mean_and_cov(res.samples[indburnin:],
                                    res.weights[indburnin:])
    print("\nWeighted mean +- sqrt(covariance)")
    if ndig is None:
        ndig = [3]*len(labels)
    for p in range(nparams):
        fmt = "{{:.{0}f}}".format(ndig[p]).format
        print(r"{0} = {1} +/- {2}".format(labels[p], fmt(mean[p]),
                                          fmt(np.sqrt(cov[p, p]))))
    if save:
        with open(output_dir+'Nestle-{}_sampling.txt'.format(method), "w") as f:
            f.write('#################################\n')
            f.write('####   CONFIDENCE INTERVALS   ###\n')
            f.write('#################################\n')
            f.write(' \n')
            f.write('Results of the NESTLE SAMPLING fit\n')
            f.write('----------------------------------\n ')
            f.write(' \n')
            f.write("\nWeighted mean +- sqrt(covariance)\n")
            for p in range(nparams):
                fmt = "{{:.{0}f}}".format(ndig[p]).format
                f.write(r"{0} = {1} +/- {2}\n".format(labels[p], fmt(mean[p]),
                                                      fmt(np.sqrt(cov[p, p]))))

    final_res = np.zeros([nparams,2])
    for p in range(nparams):
        final_res[p] = [mean[p], np.sqrt(cov[p, p])]

    if bins is None:
        bins = int(np.sqrt(res.samples[indburnin:].shape[0]))
        print("\nHist bins =", bins)

    if save or plot:
        show_corner_plot(res.samples[indburnin:], burnin=burnin, save=save,
                         output_dir=output_dir, mcmc_res=final_res, units=units,
                         ndig=ndig, labels_plot=labels_plot,
                         plot_name='Nestle-{}_corner_plot.pdf'.format(method),
                         labels=labels)
    if save:
        plt.savefig(output_dir+'Nestle-{}_corner.pdf'.format(method))
    if plot:
        plt.show()

    print('\nConfidence intervals')
    if save or plot:
        _ = confidence(res.samples[indburnin:], labels, cfd=cfd, bins=bins,
                       weights=res.weights[indburnin:],
                       gaussian_fit=True, verbose=True, save=False, **kwargs)

    if save:
        fn = output_dir+'Nestle-{}_confi_hist_gaussfit.pdf'
        plt.savefig(fn.format(method))

    if plot:
        plt.show()

    return final_res


def show_ultranest_results(un_object, labels, grid_param_list, dist, bins=None,
                           cfd=68.27, ndig=None, save=False, output_dir='/',
                           plot=False, col='b', units=None, labels_plot=None,
                           lbda_obs=None, spec_obs=None, spec_obs_err=None,
                           n_pts=None, title=None, figsize=(8,6), **kwargs):
    """
    Shows the results obtained with the Ultranest sampler: summary, parameters
    with errors, walk and corner plots. Returns best-fit values and
    uncertatinties.

    Parameters
    ----------
    un_object: object
        The UltraNest Sampler object returned by nested_spec_sampling.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state:
            - first all parameters related to loaded models (e.g. 'Teff', 'logg')
            - then the planet photometric radius 'R', in Jupiter radius
            - (optionally) the flux of emission lines (labels should match those \
            in the em_lines dictionary), in units of the model spectrum (times mu)
            - (optionally) the optical extinction 'Av', in mag
            - (optionally) the ratio of total to selective optical extinction 'Rv'
            - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb \
            contribution.
    grid_param_list : list of 1d numpy arrays/lists
        Should contain list/numpy 1d arrays with available grid of model
        parameters (should only contain the sampled parameters, not the models
        themselves). The models will be loaded with ``model_reader``.
    dist :  float
        Distance in parsec, used for flux scaling of the models.
    bins: int, optional
        The number of bins used to sample the posterior distributions.
    cfd: float, optional
        The confidence level given in percentage.
    ndig: tuple, opt
        Number of digits precision for each printed parameter.
    save: boolean, default: False
        If True, a pdf file is created.
    output_dir: str, optional
        The name of the output directory which contains the output files in the
        case  ``save`` is True.
    plot: bool, optional
        Whether to show the best-fit model against data.
    col: str, optional
        Color used for data points.
    lbda_obs, spec_obs, spec_obs_err: 1d ndarrays, optional
        Must be provided if plot is set to True
    n_pts: None or int, optional
        If None, models will be sampled at the same resolution as measured
        spectrum for plot. If an integer, corresponds to the number of sampled
        points (uniform sampling) between the first and last point of the
        measured spectrum.
    title: str, optional
        If plot is set to True, title of the plot.
    kwargs:
        Additional optional arguments passed to `nested_spec_sampling` - only
        used if plot is set to True and model parameters are evaluated.

    Returns
    -------
    final_res: numpy ndarray
         Best-fit parameters and uncertainties (corresponding to 68% confidence
         interval). Dimensions: nparams x 2.

    """
    un_sampler, res = un_object

    # print results
    un_sampler.print_results()

    if save or plot:
        # show run
        un_sampler.plot_run()
        if save:
            plt.savefig(output_dir+'UltraNest_run.pdf')
        if plot:
            plt.show()

        # show trace
        cornerplot
        un_sampler.plot_trace()
        if save:
            plt.savefig(output_dir+'UltraNest_trace.pdf')
        if plot:
            plt.show()

    burned_res = un_burning(res)

    # show results as table
    df = pd.DataFrame(data=res['samples'], columns=res['paramnames'])
    df.describe()

    # plot the best fit
    if save or plot:
        if lbda_obs is None or spec_obs is None or spec_obs_err is None:
            msg = "For the plot to be made, lbda_obs, spec_obs and spec_obs_err"
            msg += " must be provided"
            raise ValueError(msg)
        plt.figure(figsize=figsize)
        if title is not None:
            plt.title(title)
        plt.xlabel(r'Wavelength ($\mu$m)')
        plt.ylabel(r'$\lambda F_{\lambda}$ (W m-2)')
        plt.errorbar(x=lbda_obs, y=lbda_obs*spec_obs, yerr=spec_obs_err,
                     marker='o', ls=' ', color=col)

        if n_pts is None:
            lbda_grid = lbda_obs.copy()
        else:
            lbda_grid = np.linspace(lbda_obs[0], lbda_obs[-1], n_pts)
        band = PredictionBand(lbda_grid)

        # go through the solutions
        for params in un_sampler.results['samples']:
            # compute for each time the model
            mod = make_model_from_params(params, labels, grid_param_list, dist,
                                         lbda_obs=lbda_grid, **kwargs)
            band.add(lbda_grid*mod[1])

        band.line(color='k')
        # add 1 sigma quantile
        band.shade(color='k', alpha=0.3)
        # add wider quantile (0.01 .. 0.99)
        band.shade(q=0.49, color='gray', alpha=0.2)
        if save:
            plt.savefig(output_dir+"Best-fit_UltraNest.pdf",
                        bbox_inches='tight')

    print('\nConfidence intervals')
    if bins is None:
        bins = int(np.sqrt(burned_res[0].shape[0]))
        print("\nHist bins =", bins)
    if save or plot:
        val_max, confidenceInterval = confidence(burned_res[0], labels,
                                                 cfd=cfd, bins=bins,
                                                 weights=burned_res[1],
                                                 gaussian_fit=False,
                                                 verbose=True, save=save)
        if save:
            plt.savefig(output_dir+'UltraNest_confi_hist.pdf')
        if plot:
            plt.show()
        mu, sig = confidence(burned_res[0], labels, cfd=cfd, bins=bins,
                             weights=burned_res[1], gaussian_fit=True,
                             verbose=True, save=save)
        if save:
            plt.savefig(output_dir+'UltraNest_confi_hist_gaussfit.pdf')
        if plot:
            plt.show()

    nparams = len(labels)
    if ndig is None:
        ndig = [3]*len(labels)
    for p in range(nparams):
        fmt = "{{:.{0}f}}".format(ndig[p]).format
        print(r"{0} = {1} +/- {2}".format(labels[p], fmt(mu[p]),
                                          fmt(np.sqrt(sig[p]))))
    if save:
        with open(output_dir+'UltraNest_sampling.txt', "w") as f:
            f.write('#################################\n')
            f.write('####   CONFIDENCE INTERVALS   ###\n')
            f.write('#################################\n')
            f.write(' \n')
            f.write('Results of the NESTED SAMPLING (UltraNest)\n')
            f.write('------------------------------------------\n ')
            f.write(' \n')
            f.write("\nMost likely value [{}% confidence interval]\n".format(cfd))
            for p in range(nparams):
                fmt = "{{:.{0}f}}".format(ndig[p]).format
                line = r"{0} = {1} [{2}, {3}] \n"
                f.write(line.format(labels[p], fmt(val_max[labels[p]]),
                                    fmt(confidenceInterval[labels[p]][0]),
                                    fmt(confidenceInterval[labels[p]][1])))
            f.write("\nGaussian fit results: Weighted mean +- sqrt(covariance) \n")
            for p in range(nparams):
                fmt = "{{:.{0}f}}".format(ndig[p]).format
                f.write(r"{0} = {1} +/- {2}\n".format(labels[p], fmt(mu[p]),
                                                      fmt(sig[p])))


    if plot or save:
        # show corner plot
        show_corner_plot(burned_res[0], save=save, units=units, ndig=ndig,
                         mcmc_res=(val_max, confidenceInterval),
                         labels_plot=labels_plot, output_dir=output_dir,
                         plot_name='UltraNest-corner_plot.pdf',
                         labels=labels, weights=burned_res[1])
        if plot:
            plt.show()

    final_res = val_max, confidenceInterval

    return final_res


#! /usr/bin/env python

"""
Module with utility functions to the nested sampling for parameter estimation.
"""

__author__ = 'V. Christiaens'
__all__ = ['un_burning']

import numpy as np

def un_burning(res, logger=None):
    """
    Automatic burning of UltraNest chain based on cumulated sum of weights 
    (as implemented in UltraNest's cornerplot).
     
    Note: this function is necessary to be able to make corner plots showing
    units after best estimates, as ultranest's cornerplots does not feature 
    that option and does burning+corner plot together.
    
    Parameters
    ----------
    res: UltraNest result object
        The UltraNest result.
        
    Returns
    -------
    burned_res: tuple of 2 numpy nd array
        The burned UltraNest chain and associated weights
    """
    
    paramnames = res['paramnames']
    data = np.array(res['weighted_samples']['points'])
    weights = np.array(res['weighted_samples']['weights'])
    cumsumweights = np.cumsum(weights)

    mask = cumsumweights > 1e-4

    if mask.sum() == 1:
        if logger is not None:
            warn = 'Posterior is still concentrated in a single point:'
            for i, p in enumerate(paramnames):
                v = res['samples'][mask,i]
                warn += "\n" + '    %-20s: %s' % (p, v)

            logger.warning(warn)
            logger.info('Try running longer.')
        return
    
    burned_res = (data[mask,:], weights[mask])
    
    return burned_res


#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module with configuration parameters, timing functions and multiprocessing 
utilities (inspired from ``VIP``).
"""

__author__ = 'Valentin Christiaens, C. A. Gomez Gonzalez'

__all__ = ['time_ini',
           'timing',
           'time_fin',
           'sep',
           'figsize',
           'figdpi']


from datetime import datetime
import itertools as itt
import multiprocessing
import os
import sys

sep = '―' * 80
figsize = (8, 5)
figdpi = 100


def time_ini(verbose=True):
    """
    Set and print the time at which the script started.

    Returns
    -------
    start_time : string
        Starting time.

    """
    start_time = datetime.now()
    if verbose:
        print(sep)
        print("Starting time: " + start_time.strftime("%Y-%m-%d %H:%M:%S"))
        print(sep)
    return start_time


def timing(start_time):
    """
    Print the execution time of a script.

    It requires the initialization  with the function time_ini().
    """
    print("Running time:  " + str(datetime.now()-start_time))
    print(sep)


def time_fin(start_time):
    """
    Return the execution time of a script.

    It requires the initialization  with the function time_ini().
    """
    return str(datetime.now()-start_time)


class Progressbar(object):
    """ Show progress bars. Supports multiple backends.

    Examples
    --------
    .. code:: python

        from vip_hci.var import Progressbar
        Progressbar.backend = "tqdm"

        from time import sleep

        for i in Progressbar(range(50)):
            sleep(0.02)

        # or:

        bar = Progressbar(total=50):
        for i in range(50):
            sleep(0.02)
            bar.update()

        # Progressbar can be disabled globally using
        Progressbar.backend = "hide"

        # or locally using the ``verbose`` keyword:
        Progressbar(iterable, verbose=False)

    Notes
    -----
    - `leave` keyword is natively supported by tqdm, support could be added to
      other backends too?

    """
    backend = "pyprind"

    def __new__(cls, iterable=None, desc=None, total=None, leave=True,
                backend=None, verbose=True):
        if backend is None:
            backend = Progressbar.backend

        if not verbose:
            backend = "hide"

        if backend == "tqdm":
            from tqdm import tqdm
            return tqdm(iterable=iterable, desc=desc, total=total, leave=leave,
                        ascii=True, ncols=80, file=sys.stdout,
                        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed"
                                   "}<{remaining}{postfix}]") # remove rate_fmt
        elif backend == "tqdm_notebook":
            from tqdm import tqdm_notebook
            return tqdm_notebook(iterable=iterable, desc=desc, total=total,
                                 leave=leave)
        elif backend == "pyprind":
            from pyprind import ProgBar, prog_bar
            ProgBar._adjust_width = lambda self: None  # keep constant width
            if iterable is None:
                return ProgBar(total, title=desc, stream=1)
            else:
                return prog_bar(iterable, title=desc, stream=1,
                                iterations=total)
        elif backend == "hide":
            return NoProgressbar(iterable=iterable)
        else:
            raise NotImplementedError("unknown backend")

    def set(b):
        Progressbar.backend = b
        
        
class NoProgressbar(object):
    """ Wraps an ``iterable`` to behave like ``Progressbar``, but without
    producing output.
    """
    def __init__(self, iterable=None):
        self.iterable = iterable

    def __iter__(self):
        return self.iterable.__iter__()

    def __next__(self):
        return self.iterable.__next__()

    def __getattr__(self, key):
        return self.iterable.key

    def update(self):
        pass
    
    
def eval_func_tuple(f_args):
    """ Takes a tuple of a function and args, evaluates and returns result"""
    return f_args[0](*f_args[1:])


class FixedObj(object):
    def __init__(self, v):
        self.v = v


def iterable(v):
    """ Helper function for ``pool_map``: prevents the argument from being
    wrapped in ``itertools.repeat()``.

    Examples
    --------
    .. code-block:: python

        # we have a worker function whic processes a word:

        def worker(word, method):
            # ...

        # we want to process these words in parallel fasion:
        words = ["lorem", "ipsum", "esse", "ea", "eiusmod"]

        # but all with
        method = 1

        # we then would use
        pool_map(3, worker, iterable(words), method)

        # this results in calling
        #
        # worker(words[0], 1)
        # worker(words[1], 1)
        # worker(words[2], 1)
        # ...
    """
    return FixedObj(v)


def pool_map(nproc, fkt, *args, **kwargs):
    """
    Abstraction layer for multiprocessing. When ``nproc=1``, the builtin
    ``map()`` is used. For ``nproc>1`` a ``multiprocessing.Pool`` is created.

    Parameters
    ----------
    nproc : int
        Number of processes to use.
    fkt : callable
        The function to be called with each ``*args``
    *args : function arguments
        Arguments passed to ``fkt`` By default, ``itertools.repeat`` is applied
        on all the arguments, except when you wrap the argument in
        ``iterable()``.
    msg : str or None, optional
        Description to be displayed.
    progressbar_single : bool, optional
        Display a progress bar when single-processing is used. Defaults to
        ``False``.
    verbose : bool, optional
        Show more output. Also disables the progress bar when set to ``False``.

    Returns
    -------
    res : list
        A list with the results.

    """    
    msg = kwargs.get("msg", None)
    verbose = kwargs.get("verbose", True)
    progressbar_single = kwargs.get("progressbar_single", False)
    _generator = kwargs.get("_generator", False)  # not exposed in docstring

    args_r = [a.v if isinstance(a, FixedObj) else itt.repeat(a) for a in args]
    z = zip(itt.repeat(fkt), *args_r)

    if nproc == 1:
        if progressbar_single:
            total = len([a.v for a in args if isinstance(a, FixedObj)][0])
            z = Progressbar(z, desc=msg, verbose=verbose, total=total)
        res = map(eval_func_tuple, z)
        if not _generator:
            res = list(res)
    else:
        multiprocessing.set_start_method('fork', force=True)
        from multiprocessing import Pool
        # deactivate multithreading
        os.environ["MKL_NUM_THREADS"] = "1"
        os.environ["NUMEXPR_NUM_THREADS"] = "1"
        os.environ["OMP_NUM_THREADS"] = "1"

        if verbose and msg is not None:
            print("{} with {} processes".format(msg, nproc))
        pool = Pool(processes=nproc)
        if _generator:
            res = pool.imap(eval_func_tuple, z)
        else:
            res = pool.map(eval_func_tuple, z)
        pool.close()
        pool.join()

        # reactivate multithreading
        ncpus = multiprocessing.cpu_count()
        os.environ["MKL_NUM_THREADS"] = str(ncpus)
        os.environ["NUMEXPR_NUM_THREADS"] = str(ncpus)
        os.environ["OMP_NUM_THREADS"] = str(ncpus)

    return res


def pool_imap(nproc, fkt, *args, **kwargs):
    """
    Generator version of ``pool_map``. Useful when showing a progress bar for
    multiprocessing (see examples).

    Parameters
    ----------
    nproc : int
        Number of processes to use.
    fkt : callable
        The function to be called with each ``*args``
    *args : function arguments
        Arguments passed to ``fkt``
    msg : str or None, optional
        Description to be displayed.
    progressbar_single : bool, optional
        Display a progress bar when single-processing is used. Defaults to
        ``True``.
    verbose : bool, optional
        Show more output. Also disables the progress bar when set to ``False``.

    Examples
    --------

    .. code-block:: python

        # using pool_map

        res = pool_map(2, my_worker_function, *args)

        # using pool_imap with a progessbar:

        res = list(Progressbar(pool_imap(2, my_worker_function, *args)))

    """
    kwargs["_generator"] = True
    return pool_map(nproc, fkt, *args, **kwargs)


def repeat(*args):
    """
    Applies ``itertools.repeat`` to every ``args``.


    Examples
    --------

    # instead of using

    import itertools as itt

    my_fkt(itt.repeat(a), itt.repeat(b), itt.repeat(c), d, itt.repeat(e))

    # you could use `repeat`:

    my_fkt(*repeat(a, b, c), d, *repeat(e))

    """
    return [itt.repeat(a) for a in args]



#! /usr/bin/env python

"""
Module for simplex or grid search of best fit spectrum in a template library.
"""

__author__ = 'V. Christiaens'
__all__ = ['best_fit_tmp',
           'get_chi']

from datetime import datetime
from multiprocessing import cpu_count
import numpy as np
import os
from scipy.optimize import minimize
from .config import time_ini, timing, time_fin, pool_map, iterable
from .chi import gof_scal
from .model_resampling import resample_model
from .utils_spec import extinction, find_nearest
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


def get_chi(lbda_obs, spec_obs, err_obs, tmp_name, tmp_reader, 
            search_mode='simplex', lambda_scal=None, scale_range=(0.1,10,0.01), 
            ext_range=None, dlbda_obs=None, instru_corr=None, instru_res=None, 
            instru_idx=None, use_weights=True, filter_reader=None, 
            simplex_options=None, red_chi2=True, remove_nan=False, 
            force_continue=False, min_npts=1, verbose=False, **kwargs):
    """ Routine calculating chi^2, optimal scaling factor and optimal 
    extinction for a given template spectrum to match an observed spectrum.
    
    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelengths of observed spectrum. If several instruments were used, the 
        wavelengths should be ordered per instrument, not necessarily as 
        monotonically increasing wavelength. Hereafter, :math:`n_{ch}` is the 
        length of ``lbda_obs``.
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of ``lbda_obs``. Should have a length
        of :math:`n_{ch}`.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. The array (list) can have either
        a length of :math:`n_{ch}`, or a shape of :math:`(2,n_{ch})` for lower 
        (first column) and upper (second column) uncertainties provided.
    tmp_name :  str
        Template spectrum filename.
    tmp_reader : python routine
        External routine that reads a model file and returns a 3D numpy array, 
        where the first column corresponds to wavelengths, the second 
        contains flux values, and the third the uncertainties on the flux.
    search_mode: str, opt {'simplex', 'grid'}
        How is the best fit template found? Simplex or grid search.
    lambda_scal: float, optional
        Wavelength where a first scaling will be performed between template
        and observed spectra. If not provided, the middle wavelength of the 
        osberved spectra will be considered.
    scale_range: tuple, opt
        If grid search, this parameter should be provided as a tuple of 3 
        floats: lower limit, upper limit and step of the grid search for the 
        scaling factor to be applied AFTER the first rough scaling (i.e.
        scale_range should always encompass 1).
    ext_range : tuple or None, opt
        - If None: differential extinction is not considered as a free parameter. 
        - If a tuple: it should contain 2 floats (for simplex \
        ``search_mode``) or 3 floats (for grid search ``search_mode``) \
        corresponding to the lower limit, upper limit (and step for the grid \
        search). For the simplex search, the lower and upper limits are used \
        to set a chi square of infinity outside of the range.
    dlbda_obs : numpy 1d ndarray or list, optional
        Respective spectral channel width or FWHM of the photometric filters 
        used for each point of the observed spectrum. This vector is used to 
        infer which part(s) of a combined spectro+photometric spectrum should 
        involve convolution+subsampling (model resolution higher than 
        measurements), interpolation (the opposite), or convolution by the 
        transmission curve of a photometric filter. If not provided, it will be 
        inferred from the difference between consecutive lbda_obs points (i.e. 
        inaccurate for a combined spectrum). It must be provided IF one wants to 
        weigh each measurement based on the spectral resolution of each 
        instrument (as in [OLO16]_), through the ``use_weights`` argument.
    instru_corr : numpy 2d ndarray, optional
        Spectral correlation between post-processed images in which the 
        spectrum is measured. It is specific to the instrument, PSF subtraction 
        algorithm and radial separation of the companion from the central star.
        Can be computed using ``special.spec_corr.spectral_correlation``. In 
        case of a spectrum obtained with different instruments, it is 
        recommended to construct the final spectral_correlation matrix with
        ``special.spec_corr.combine_corrs``. If ``instru_corr`` is not provided, 
        the uncertainties in each spectral channel will be considered 
        independent. See [GRE16]_ for more details.
    instru_res : float or list of floats/strings, optional
        The mean instrumental resolving power(s) OR filter names. This is 
        used to convolve the model spectrum. If several instruments are used, 
        provide a list of resolving power values / filter names, one for 
        each instrument used.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to the number of instruments 
        (:math:`n_{ins}`). Zero for points that don't correspond to any of the 
        ``instru_res`` values provided, and i in :math:`[1,n_{ins}]` for points
        associated to instru_res[i-1]. This parameter must be provided if the 
        spectrum consists of points obtained with different instruments.
    use_weights: bool, optional
        For the likelihood calculation, whether to weigh each point of the 
        spectrum based on the spectral resolution or bandwidth of photometric
        filters used. Weights will be proportional to ``dlbda_obs/lbda_obs`` if 
        ``dlbda_obs`` is provided, or set to 1 for all points otherwise.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_res, the default file reader will be used. 
        It assumes the following format for the files:
            
        - first row contains headers (titles of each column)
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header \
        key name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer \
        or "lambda(nm)" for nanometer. Note: only what is in parentheses \
        matters for the units.
     
    red_chi2: bool, optional
        Whether to compute the reduced chi square. If False, considers chi^2.
    remove_nan: bool, optional
        Whether to remove NaN values from template spectrum BEFORE resampling
        to the wavelength sampling of the observed spectrum. Whether it is set
        to True or False, a check is made just before chi^2 is calculated 
        (after resampling), and only non-NaN values will be considered.
    simplex_options: dict, optional
        The ``scipy.optimize.minimize`` simplex (Nelder-Mead) options.
    force_continue: bool, optional
        In case of issue with the fit, whether to continue regardless (this may
        be useful in an uneven spectral library, where some templates have too
        few points for the fit to be performed).
    verbose: str, optional
        Whether to print more information when fit fails. 
    min_npts: int or None, optional
        Iinimum number of (resampled) points to consider a template spectrum 
        valid in the minimization search. A Nan value will be returned for chi 
        if the condition is not met.
    **kwargs: optional
        Other optional arguments to the ``scipy.optimize.minimize`` function.
        
    Returns
    -------
    best_chi: float
        goodness of fit scored by the template
    best_scal:
        best-fit scaling factor for the considered template
    best_ext:
        best-fit optical extinction for the considered template

    Note
    ----
    If several filter filenames are provided in ``instru_res``, the filter files
    must all have the same format and wavelength units (for reading by the same
    ``filter_reader`` snippet or default function). 
        
    """
    # read template spectrum
    try:
        lbda_tmp, spec_tmp, spec_tmp_err = tmp_reader(tmp_name, 
                                                      verbose=verbose>1)
    except:
        msg = "{} could not be opened. Corrupt file?".format(tmp_name)
        if force_continue:
            if verbose:
                print(msg)
            return np.inf, 1, 0, 1
        else:
            raise ValueError(msg)
            
    # look for any nan and replace
    if remove_nan:
        if np.isnan(spec_tmp).any() or np.isnan(spec_tmp_err).any():
            bad_idx = np.where(np.isnan(spec_tmp))[0]
            #bad_idx2 = np.where(np.isnan(spec_tmp_err))[0]
            #all_bad = np.concatenate((bad_idx1,bad_idx2))
            nch = len(lbda_tmp)
            new_lbda = [lbda_tmp[i] for i in range(nch) if i not in bad_idx]
            new_spec = [spec_tmp[i] for i in range(nch) if i not in bad_idx]
            new_err = [spec_tmp_err[i] for i in range(nch) if i not in bad_idx]
            lbda_tmp = np.array(new_lbda)
            spec_tmp = np.array(new_spec)
            spec_tmp_err = np.array(new_err)
        
    # don't consider template spectra whose range is smaller than observed one
    if lbda_obs[0] < lbda_tmp[0] or lbda_obs[-1] > lbda_tmp[-1]:
        msg = "Wavelength range of template {} ({:.2f}, {:.2f})mu too short"
        msg+= " compared to that of observed spectrum ({:.2f}, {:.2f})mu"
        if force_continue:
            if verbose:
                print(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1],
                                        lbda_obs[0],lbda_obs[-1]))
            return np.inf, 1, 0, len(lbda_tmp)-2
        else:
            raise ValueError(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1],
                                        lbda_obs[0],lbda_obs[-1]))
    
    # try to resample tmp as observed spectrum - just used to raise error early
    try:
        _, spec_res = resample_model(lbda_obs, lbda_tmp, spec_tmp, 
                                     dlbda_obs=dlbda_obs, 
                                     instru_res=instru_res, 
                                     instru_idx=instru_idx,
                                     filter_reader=filter_reader)
    except:
        msg = "Issue with resampling of template {}. Does the wavelength "
        msg+= "range extend far enough ({:.2f}, {:.2f})mu?"
        if force_continue:
            if verbose:
                print(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1]))
            return np.inf, 1, 0, len(lbda_tmp)-2
        else:
            raise ValueError(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1]))
    
    # first rescaling fac
    if not lambda_scal:
        lambda_scal = (lbda_obs[0]+lbda_obs[-1])/2
    idx_cen = find_nearest(lbda_obs, lambda_scal)
    idx_tmp = find_nearest(lbda_tmp, lambda_scal)
    scal_fac = spec_obs[idx_cen]/spec_tmp[idx_tmp]
    spec_tmp*=scal_fac
    #spec_tmp_err*=scal_fac
    
    # EDIT: Don't combine observed and template uncertainties;
    # the best fit would be the most noisy tmp of the library!)
    #err_obs = np.sqrt(np.power(spec_tmp_err,2)+np.power(err_obs,2))
    
    
    # only consider non-zero and non-nan values for chi^2 calculation
    all_conds = np.where(np.isfinite(spec_res))[0]
    # cond2 = np.where(np.isfinite(err_obs))[0] 
    # cond3 = np.where(spec_tmp>0)[0]
    # all_conds = np.sort(np.unique(np.concatenate((cond1,cond2,cond3))))
    ngood_ch = len(all_conds)
    #good_ch = (all_conds,)
    # lbda_obs = lbda_obs[good_ch]
    # spec_obs = spec_obs[good_ch]
    # err_obs = err_obs[good_ch]
    #lbda_tmp = lbda_tmp[good_ch]
    #spec_tmp = spec_tmp[good_ch]
    
    n_dof = ngood_ch-1-(ext_range is not None)
    if n_dof <= 0:
        msg = "Not enough dof with remaining points of template spectrum {}"
        if force_continue:
            if verbose:
                print(msg.format(tmp_name))
            return np.inf, 1, 0, n_dof
        else:
            raise ValueError(msg.format(tmp_name))
    
    best_chi = np.inf
    best_scal = np.nan
    best_ext = np.nan
    if ngood_ch < min_npts:
        msg = "Unsufficient number of good points ({} < {}). Tmp discarded."
        if verbose:
            print(msg.format(ngood_ch,min_npts))
        return best_chi, best_scal, best_ext, n_dof
    
    # simplex search
    if search_mode == 'simplex':
        if simplex_options is None:
            simplex_options = {'xatol': 1e-6, 'fatol': 1e-6, 'maxiter': 1000,
                               'maxfev': 5000}
        if not ext_range:
            p = (1,)
        else:
            AV_ini = (ext_range[0]+ext_range[1])/2
            p = (1,AV_ini)
        
        try:
            res = minimize(gof_scal, p, args=(lbda_obs, spec_obs, err_obs, 
                                              lbda_tmp, spec_tmp, dlbda_obs, 
                                              instru_corr, instru_res, 
                                              instru_idx, use_weights, 
                                              filter_reader, ext_range),
                           method='Nelder-Mead', options=simplex_options, 
                           **kwargs)
        except:
            msg = "Issue with simplex minimization for template {}. "
            msg+= "Try grid search?"
            if force_continue:
                if verbose:
                    print(msg.format(tmp_name))
                return np.inf, 1, 0, n_dof
            else:
                raise ValueError(msg.format(tmp_name))
        best_chi = res.fun
        if not ext_range:
            best_scal = res.x
            best_ext = 0
        else:
            best_scal, best_ext = res.x
        if np.isfinite(best_scal):
            best_scal*=scal_fac
        
    # or grid search        
    elif search_mode == 'grid':
        test_scale = np.arange(scale_range[0], scale_range[1], scale_range[2])
        n_test = len(test_scale)
        if ext_range is None:
            test_ext = np.array([0])
            n_ext = 1
        elif isinstance(ext_range, tuple) and len(ext_range)==3:
            test_ext = np.arange(ext_range[0], ext_range[1], ext_range[2])
            n_ext = len(test_ext)
        else:
            raise TypeError("ext_range can only be None or tuple of length 3")

        chi = np.zeros([n_test,n_ext])
        
        for cc, scal in enumerate(test_scale):
            for ee, AV in enumerate(test_ext):
                p = (scal,AV)
                chi[cc,ee] = gof_scal(p, lbda_obs, spec_obs, err_obs, lbda_tmp, 
                                      spec_tmp, dlbda_obs=dlbda_obs, 
                                      instru_corr=instru_corr, 
                                      instru_res=instru_res, 
                                      instru_idx=instru_idx, 
                                      use_weights=use_weights,
                                      filter_reader=filter_reader,
                                      ext_range=ext_range)
        try:
            best_chi = np.nanmin(chi)
            best_idx = np.nanargmin(chi)
            best_idx = np.unravel_index(best_idx,chi.shape)
            best_scal = test_scale[best_idx[0]]*scal_fac
            best_ext = test_ext[best_idx[1]]
        except:
            if force_continue:
                return best_chi, best_scal, best_ext, n_dof
            else:
                msg = "Issue with grid search minimization for template {}. "
                print(msg.format(tmp_name))
                import pdb
                pdb.set_trace()
    
    else:
        msg = "Search mode not recognised. Should be 'simplex' or 'grid'."
        raise TypeError(msg)
    
    if red_chi2:
        best_chi /= n_dof
        
    
    return best_chi, best_scal, best_ext, n_dof



def best_fit_tmp(lbda_obs, spec_obs, err_obs, tmp_reader, search_mode='simplex',
                 n_best=1, lambda_scal=None, scale_range=(0.1,10,0.01), 
                 ext_range=None, simplex_options=None, dlbda_obs=None, 
                 instru_corr=None, instru_res=None, instru_idx=None, 
                 filter_reader=None, lib_dir='tmp_lib/', tmp_endswith='.fits', 
                 red_chi2=True, remove_nan=False, nproc=1, verbosity=0, 
                 force_continue=False, min_npts=1, **kwargs):
    """ Finds the best fit template spectrum to a given observed spectrum, 
    within a spectral library.  By default, a single free parameter is 
    considered: the scaling factor of the spectrum. A first automatic scaling 
    is performed by comparing the flux of the observed and template spectra at 
    lambda_scal. Then a more refined scaling is performed, either through 
    simplex or grid search (within scale_range).
    If fit_extinction is set to True, the exctinction is also considered as a 
    free parameter.
    
    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelengths of observed spectrum. If several instruments were used, the 
        wavelengths should be ordered per instrument, not necessarily as 
        monotonically increasing wavelength. Hereafter, :math:`n_{ch}` is the 
        length of ``lbda_obs``.
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of ``lbda_obs``. Should have a length
        of :math:`n_{ch}`.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. The array (list) can have either
        a length of :math:`n_{ch}`, or a shape of :math:`(2,n_{ch})` for lower 
        (first column) and upper (second column) uncertainties provided.
    tmp_reader : python routine
        External routine that reads a model file and returns a 3D numpy array, 
        where the first column corresponds to wavelengths, the second 
        contains flux values, and the third the uncertainties on the flux.
    search_mode: str, optional, {'simplex','grid'}
        How is the best fit template found? Simplex or grid search.
    n_best: int, optional
        Number of best templates to be returned (default: 1)
    lambda_scal: float, optional
        Wavelength where a first scaling will be performed between template
        and observed spectra. If not provided, the middle wavelength of the 
        osberved spectra will be considered.
    scale_range: tuple, opt
        If grid search, this parameter should be provided as a tuple of 3 
        floats: lower limit, upper limit and step of the grid search for the 
        scaling factor to be applied AFTER the first rough scaling (i.e.
        scale_range should always encompass 1).
    ext_range : tuple or None, opt
        - If None: differential extinction is not considered as a free parameter. 
        - If a tuple: it should contain 2 floats (for simplex \
        ``search_mode``) or 3 floats (for grid search ``search_mode``) \
        corresponding to the lower limit, upper limit (and step for the grid \
        search). For the simplex search, the lower and upper limits are used \
        to set a chi square of infinity outside of the range.
    dlbda_obs : numpy 1d ndarray or list, optional
        Respective spectral channel width or FWHM of the photometric filters 
        used for each point of the observed spectrum. This vector is used to 
        infer which part(s) of a combined spectro+photometric spectrum should 
        involve convolution+subsampling (model resolution higher than 
        measurements), interpolation (the opposite), or convolution by the 
        transmission curve of a photometric filter. If not provided, it will be 
        inferred from the difference between consecutive lbda_obs points (i.e. 
        inaccurate for a combined spectrum). It must be provided IF one wants to 
        weigh each measurement based on the spectral resolution of each 
        instrument (as in [OLO16]_), through the ``use_weights`` argument.
    instru_corr : numpy 2d ndarray, optional
        Spectral correlation between post-processed images in which the 
        spectrum is measured. It is specific to the instrument, PSF subtraction 
        algorithm and radial separation of the companion from the central star.
        Can be computed using ``special.spec_corr.spectral_correlation``. In 
        case of a spectrum obtained with different instruments, it is 
        recommended to construct the final spectral_correlation matrix with
        ``special.spec_corr.combine_corrs``. If ``instru_corr`` is not provided, 
        the uncertainties in each spectral channel will be considered 
        independent. See [GRE16]_ for more details.
    instru_res : float or list of floats/strings, optional
        The mean instrumental resolving power(s) OR filter names. This is 
        used to convolve the model spectrum. If several instruments are used, 
        provide a list of resolving power values / filter names, one for 
        each instrument used.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to the number of instruments 
        (:math:`n_{ins}`). Zero for points that don't correspond to any of the 
        ``instru_res`` values provided, and i in :math:`[1,n_{ins}]` for points
        associated to instru_res[i-1]. This parameter must be provided if the 
        spectrum consists of points obtained with different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_res, the default file reader will be used. 
        It assumes the following format for the files:
            
        - first row contains headers (titles of each column)
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header \
        key name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer \
        or "lambda(nm)" for nanometer. Note: only what is in parentheses \
        matters for the units.
     
    simplex_options: dict, optional
        The scipy.optimize.minimize simplex (Nelder-Mead) options.
    red_chi2: bool, optional
        Whether to compute the reduced chi square. If False, considers chi^2.
    remove_nan: bool, optional
        Whether to remove NaN values from template spectrum BEFORE resampling
        to the wavelength sampling of the observed spectrum. Whether it is set
        to True or False, a check is made just before chi^2 is calculated 
        (after resampling), and only non-NaN values will be considered.
    nproc: None or int, optional
        The number of processes to use for parallelization. If set to None, 
        will automatically use half of the available CPUs on the machine.
    verbosity: 0, 1 or 2, optional
        Verbosity level. 0 for no output and 2 for full information.
    force_continue: bool, optional
        In case of issue with the fit, whether to continue regardless (this may
        be useful in an uneven spectral library, where some templates have too
        few points for the fit to be performed).
    min_npts: int or None, optional
        Minimum number of (resampled) points to consider a 
        template spectrum valid in the minimization search.
    **kwargs: optional
        Optional arguments to the scipy.optimize.minimize function
         
    Returns
    -------
    final_tmpname: tuple of n_best str
        Best-fit template filenames
    final_tmp: tuple of n_best 3D numpy array
        Best-fit template spectra (3D: lbda+spec+spec_err)
    final_chi: 1D numpy array of length n_best
        Best-fit template chi^2
    final_params: 2D numpy array (2xn_best)
        Best-fit parameters (optimal scaling and optical extinction). Note if 
        extinction is not fitted, optimal AV will be set to 0.
        
    """
    # create list of template filenames
    tmp_filelist = [x for x in os.listdir(lib_dir) if x.endswith(tmp_endswith)]
    n_tmp = len(tmp_filelist)
    
    if verbosity > 0:
        start_time = time_ini()
        int_time = start_time
        print("{:.0f} template spectra will be tested.".format(n_tmp))
    
    chi = np.ones(n_tmp)
    scal = np.ones_like(chi)
    ext = np.zeros_like(chi)
    n_dof =  np.ones_like(chi)
    
    if nproc is None:
        nproc = cpu_count()//2
        if verbosity:
            print("{:.0f} CPUs will be used".format(nproc))

    if verbosity:
        print("****************************************\n")
    
    if nproc == 1:
        for tt in range(n_tmp):
            if verbosity>1 and search_mode=='simplex':
                print('Nelder-Mead minimization is running...')
            res = get_chi(lbda_obs, spec_obs, err_obs, tmp_filelist[tt], 
                          tmp_reader, search_mode=search_mode, 
                          scale_range=scale_range, ext_range=ext_range,
                          lambda_scal=lambda_scal, dlbda_obs=dlbda_obs, 
                          instru_corr=instru_corr, instru_res=instru_res, 
                          instru_idx=instru_idx, filter_reader=filter_reader,
                          simplex_options=simplex_options, red_chi2=red_chi2,
                          remove_nan=remove_nan, force_continue=force_continue,
                          verbose=verbosity, min_npts=min_npts, **kwargs)
            chi[tt], scal[tt], ext[tt], n_dof[tt] = res
            
            shortname = tmp_filelist[tt][:-len(tmp_endswith)]
            if not np.isfinite(chi[tt]):
                if verbosity>0:
                    msg_err = "{:.0f}/{:.0f} ({}) FAILED"
                    if np.isnan(chi[tt]):
                        msg_err += " (simplex did not converge)"
                    print(msg_err.format(tt+1, n_tmp, tmp_filelist[tt]))
            else:
                if verbosity > 0 and tt==0:
                    msg = "{:.0f}/{:.0f}: {}, chi_r^2 = {:.1f}, ndof={:.0f}"
                    if verbosity>1:
                        msg+=", done in {}s."
                        indiv_time = time_fin(start_time)
                        print(msg.format(tt+1, n_tmp, shortname, chi[tt], 
                                         n_dof[tt], indiv_time))
                    else:
                        print(msg.format(tt+1, n_tmp, shortname, chi[tt], n_dof[tt]))
                    now = datetime.now()
                    delta_t = now.timestamp()-start_time.timestamp()
                    tot_time = np.ceil(n_tmp*delta_t/60)
                    msg = "Based on the first fit, it may take ~{:.0f}min to"
                    msg += " test the whole library \n"
                    print(msg.format(tot_time))
                    int_time = time_ini(verbose=False)
                elif verbosity > 0:
                    msg = "{:.0f}/{:.0f}: {}, chi_r^2 = {:.1f}, ndof={:.0f}"
                    if verbosity>1:
                        msg+=" done in {}s."                    
                        indiv_time = time_fin(int_time)
                        int_time = time_ini(verbose=False)
                        print(msg.format(tt+1, n_tmp, shortname, chi[tt], 
                                         n_dof[tt], indiv_time))
                    else:
                        print(msg.format(tt+1, n_tmp, shortname, chi[tt],
                                         n_dof[tt]))
                        

    else:
        res = pool_map(nproc, get_chi, lbda_obs, spec_obs, err_obs, 
                       iterable(tmp_filelist), tmp_reader, search_mode, 
                       lambda_scal, scale_range, ext_range, dlbda_obs, 
                       instru_corr, instru_res, instru_idx, filter_reader,
                       simplex_options, red_chi2, remove_nan, force_continue,
                       verbosity, min_npts)

        res = np.array(res, dtype=object)
        chi = res[:,0]
        scal = res[:,1]
        ext = res[:,2]
        n_dof = res[:,3]
        
    n_success = np.sum(np.isfinite(chi))
        
    if verbosity > 0:
        print("****************************************\n")
        msg = "{:.0f}/{:.0f} template spectra were fitted. \n"
        print(msg.format(n_success, n_tmp))
        timing(start_time)
        
    return best_n_tmp(chi, scal, ext, n_dof, tmp_filelist, tmp_reader, 
                      n_best=n_best, verbose=verbosity)
    
    
def best_n_tmp(chi, scal, ext, n_dof, tmp_filelist, tmp_reader, n_best=1, 
               verbose=False):
    """
    Routine returning the n best template spectra.

    Returns
    -------
    final_tmpname: tuple of n_best str
        Best-fit template filenames
    final_tmp: tuple of n_best 3D numpy array
        Best-fit template spectra (3D: lbda+spec+spec_err)
    final_chi: 1D numpy array of length n_best
        Best-fit template chi^2
    final_params: 2D numpy array (2xn_best)
        Best-fit parameters (optimal scaling and optical extinction). Note if 
        extinction is not fitted, optimal AV will be set to 0.
        
    """
    # sort from best to worst match
    order = np.argsort(chi)
    sort_chi = chi[order]
    sort_scal = scal[order]
    sort_ext = ext[order]
    sort_ndof = n_dof[order]
    sort_filelist = [tmp_filelist[i] for i in order]
    
    if verbose:
        print("best chi: ", sort_chi[:n_best])
        print("best scale fac: ", sort_scal[:n_best])
        print("n_dof: ", sort_ndof[:n_best])
    # take the n_best first ones
    best_tmp = []
    for n in range(n_best):
        lbda_tmp, spec_tmp, spec_tmp_err = tmp_reader(sort_filelist[n])
        Albdas = extinction(lbda_tmp,abs(sort_ext[n]))
        extinc_fac = np.power(10.,-Albdas/2.5)
        if sort_ext[n]>0:
            final_scal = sort_scal[n]*extinc_fac
        elif sort_ext[n]<0:
            final_scal = sort_scal[n]/extinc_fac
        else:
            final_scal = sort_scal[n]
        best_tmp.append(np.array([lbda_tmp, spec_tmp*final_scal, 
                                  spec_tmp_err*final_scal]))
        if verbose:
            msg = "The best template #{:.0f} is: {} "
            msg+="(Delta A_V={:.1f}mag)\n"
            print(msg.format(n+1, sort_filelist[n], sort_ext[n]))
            
    best_tmpname = tuple(sort_filelist[:n_best])
    best_tmp = tuple(best_tmp)
    best_params = np.array([sort_scal[:n_best],sort_ext[:n_best]])
    
    return (best_tmpname, best_tmp, sort_chi[:n_best], best_params, 
            sort_ndof[:n_best])


"""
``special`` has helping functions for the analysis of (low-res) 
spectra, including:

- fitting of input spectra to models and templates;
- mcmc sampling of model parameter space;
- nested sampling of model parameter space;
- best fit search within a template library;
- utility functions for the spectral fit.
"""

__version__ = "0.1.23"

from .config import *
from .utils_mcmc import *
from .utils_spec import *
from .fits import *
from .spec_corr import *
from .spec_indices import *
from .chi import *
from .model_resampling import *
from .mcmc_sampling import *
from .nested_sampling import *
from .template_fit import *



#! /usr/bin/env python

"""
Utility functions for spectral fitting.
"""

__author__ = 'Valentin Christiaens'
__all__ = ['akaike',
           'blackbody',
           'convert_F_units',
           'convert_F_vs_mag',
           'extinction',
           'find_nearest',
           'inject_em_line',
           'mj_from_rj_and_logg',
           'nrefrac']

import astropy.constants as c
import numpy as np
from scipy.signal import gaussian
        
def akaike(LnL, k):
    """
    Computes the Akaike Information Criterion: 2k-2ln(L),
    where k is the number of estimated parameters in the model and LnL is the 
    max ln-likelihood for the model.
    
    Parameters
    ----------
    LnL : float
        Max ln-likelihood for the considered model.
    k : int
        Number of estimated parameters in the model.

    Returns
    -------
    aic : float
        Akaike Information Criterion
        
    """
    return 2*k-2*LnL  


def blackbody(lbda, T): 
    """
    Planck function. Returns specific intensity for an input wavelength vector
    lbda (in micrometers) and a given input temperature.
    
    Parameters
    ----------
    lbda : numpy array
        1d numpy array corresponding to the wavelengths (in microns) for the 
        desired output specific intensities.
    T : float
        Temperature

    Returns
    -------
    B_lambda : float
        Specific intensity corresponding to the Planck function.
    
    """
    fac = 2*c.h.value*(c.c.value**2)/(np.power(lbda*1e-6,5))
    div = (1/(np.exp((c.h.value*c.c.value)/((lbda*1e-6)*c.k_B.value*T))-1))
    # convert from W m-3 Sr-1 to W m-2 mu-1 Sr-1 
    conv = 1e-6
    return fac*div*conv


def convert_F_units(F, lbda, in_unit='cgs', out_unit='si'):
    """
    Function to convert Flux density between [ergs s-1 cm-2 um-1], 
    [W m-2 um-1] and [Jy].

    Parameters
    ----------
    F: float or 1d array
        Flux
    lbda: float or 1d array
        Wavelength of the flux (in um)
    in_unit: str, opt, {"si", "cgs", "jy", "cgsA"} 
        Input flux units. 
        'si': W/m^2/mu; 
        'cgs': ergs/s/cm^2/mu
        'jy': janskys
        'cgsA': erg/s/cm^2/AA 
    out_unit: str, opt {"si", "cgs", "jy"}
        Output flux units.
        
    Returns
    -------
    Flux in output units.
    """

    if in_unit == 'cgs':
        new_F = (F*1e23*np.power(lbda,2))/(c.c.value*1e6) # convert to jy
    elif in_unit == 'cgsA':
        new_F = (F*1e27*np.power(lbda,2))/(c.c.value*1e6) # convert to jy
    elif in_unit == 'si':
        new_F = (F*1e26*np.power(lbda,2))/(c.c.value*1e6) # convert to jy
    elif in_unit == "jy":
        new_F=F
    else:
        msg = "in_unit not recognized, try either 'cgs', 'si' or 'jy'."
        raise TypeError(msg)        
    if out_unit == 'jy':
        return new_F
    elif out_unit == 'cgs':
        return new_F*1e-23*c.c.value*1e6/np.power(lbda,2)
    elif out_unit == 'si':
        return new_F*1e-26*c.c.value*1e6/np.power(lbda,2)
    else:
        msg = "out_unit not recognized, try either 'cgs', 'si' or 'jy'."
        raise TypeError(msg)  


def convert_F_vs_mag(value, F_0=None, band='H', system='Johnson', 
                     conversion='to_mag'):
    """
    Function to convert Flux density (in Jy) to magnitude in a given band, or 
    the opposite.

    Sources for zero points:
        * TOKUNAGA chapter on IR astronomy (from Cohen 1992)
        * UKIRT webpage: \
        (http://www.jach.hawaii.edu/UKIRT/astronomy/calib/phot_cal/conver.html)
        * van der Bliek et al. 1996 (ESO standard stars)

    Parameters
    ----------
    value: float
        Flux or magnitude to be converted.
    F_0: float, opt
        Zero-point flux. If provided will take precedence over band.
    band: str, opt
        Band of the given flux or magnitude. Choice between: {'U','B','V', 'R', 
        'I', 'J', 'H', 'K', "L", "L'", 'M', 'N', 'O'} 
        (but not for all band systems).
    system: str, opt
        Band system. Choice between: {'Johnson;,'2MASS', 'UKIRT', 'ESO'}
    conversion: str, opt
        In which sense to convert: flux to mag ('to_mag') or mag to flux 
        ('to_flux')
    
    Returns
    -------
    Converted flux or magnitude.
    """               
    
    dico_zero_pts_Jo = {'U': [0.36,1823.],
                        'B': [0.44,4130.],
                        'V': [0.55,3781.],
                        'R': [0.71,2941.],
                        'I': [0.97,2635.],
                        'J': [1.25,1603.],
                        'H': [1.60,1075.],
                        'K': [2.22,667.],
                        'L': [3.54,288.],
                        'M': [4.80,170.],
                        'N': [10.6,36.],
                        'O': [21.0,9.4]}
    dico_zero_pts_2M = {'J': [1.235,1594.],
                        'H': [1.662,1024.],
                        'K': [2.159,666.7]}
    dico_zero_pts_UK = {'V': [0.5556,3540.], # TOKUNAGA (from Cohen 1992)
                        'I': [0.9,2250.],    # UKIRT webpage
                        'J': [1.215,1630.],  # TOKUNAGA (from Cohen 1992)
                        'H': [1.654,1050.],  # TOKUNAGA (from Cohen 1992)
                        'Ks': [2.157,667.],  # TOKUNAGA (from Cohen 1992)
                        'K': [2.179,655.],   # TOKUNAGA (from Cohen 1992)                        
                        'L': [3.547,276.],   # TOKUNAGA (from Cohen 1992)  
                        "L'": [3.761,248.],  # TOKUNAGA (from Cohen 1992)                          
                        'M': [4.769,160.],   # TOKUNAGA (from Cohen 1992)  
                        '8.7': [8.756,50.],  # TOKUNAGA (from Cohen 1992)                          
                        'N': [10.472,35.3],  # TOKUNAGA (from Cohen 1992)  
                        '11.7': [11.653,28.6], # TOKUNAGA (from Cohen 1992)         
                        'Q': [20.13,9.7]}      # TOKUNAGA (from Cohen 1992)
    dico_zero_pts_ESO = {'J': [1.228,3.44e-9],  # van der Bliek 1996
                        'H': [1.651,1.21e-9],   # van der Bliek 1996
                        'K': [2.216,4.12e-10],  # van der Bliek 1996
                        "L'": [3.771,5.58e-11], # van der Bliek 1996
                        "M": [4.772,2.21e-11]}  # van der Bliek 1996 
    
    if F_0 is None:
        if system == 'Johnson' and band in dico_zero_pts_Jo:
            dico_F_0 = dico_zero_pts_Jo
        elif system == '2MASS' and band in dico_zero_pts_2M:
            dico_F_0 = dico_zero_pts_2M
        elif system == 'UKIRT' and band in dico_zero_pts_UK:
            dico_F_0 = dico_zero_pts_UK
        elif system == 'ESO' and band in dico_zero_pts_UK:
            dico_F_0 = dico_zero_pts_ESO                   
        else:
            msg = 'Combination of band name and band system not recognized.'
            raise TypeError(msg)
        F_0 = dico_F_0[band][1]
        if system == 'ESO':
            # convert from W m-2 mu-1 to Jy
            F_0 = convert_F_units(F_0, dico_F_0[band][0], in_unit='si', 
                                  out_unit='jy')
    
    if conversion == 'to_mag':
        return -2.5*np.log10(value/F_0)
    elif conversion == 'to_flux':
        return F_0*np.power(10.,-value/2.5)
    else:
        msg = "conversion not recognized, must be 'to_mag' or 'to_flux'."
        raise TypeError(msg)
        

def extinction(lbda, AV, RV=3.1):
    """
    Calculates the A(lambda) extinction for a given combination of A_V and R_V.
    If R_V is not provided, assumes an ISM value of R_V=3.1
    Uses the Cardelli et al. (1989) empirical formulas.
    
    Parameters
    ----------
    lbda : 1d np.ndarray
        Array with the wavelengths (um) for which the extinction is calculated.
    AV : float
        Extinction (mag) in the V band.
    RV : float, opt
        Reddening in the V band: R_V = A_V / E(B-V)
        
    Returns
    -------
    Albda: 1d np.ndarray
        Extinction (mag) at wavelengths lbda.
    """

    xx = 1./lbda
    yy = xx - 1.82

    a_c = np.zeros_like(xx)
    b_c = np.zeros_like(xx)

    indices = np.where(xx < 1.1)[0]

    if len(indices) > 0:
        a_c[indices] = 0.574*np.power(xx[indices], 1.61)
        b_c[indices] = -0.527*np.power(xx[indices], 1.61)

    indices = np.where(xx >= 1.1)[0]

    if len(indices) > 0:
        a_c[indices] = 1. + 0.17699*yy[indices] - 0.50447*yy[indices]**2 - \
                       0.02427*yy[indices]**3 + 0.72085*yy[indices]**4 + \
                       0.01979*yy[indices]**5 - 0.77530*yy[indices]**6 + \
                       0.32999*yy[indices]**7

        b_c[indices] = 1.41338*yy[indices] + 2.28305*yy[indices]**2 + \
                       1.07233*yy[indices]**3 - 5.38434*yy[indices]**4 - \
                       0.62251*yy[indices]**5 + 5.30260*yy[indices]**6 - \
                       2.09002*yy[indices]**7

    return AV * (a_c + b_c/RV)


def find_nearest(array, value, output='index', constraint=None, n=1):
    """
    Function to find the indices, and optionally the values, of an array's n 
    closest elements to a certain value.
    
    Parameters
    ----------
    array: 1d numpy array or list
        Array in which to check the closest element to value.
    value: float
        Value for which the algorithm searches for the n closest elements in 
        the array.
    output: str, opt {'index','value','both' }
        Set what is returned
    constraint: str, opt {None, 'ceil', 'floor'}
        If not None, will check for the closest element larger than value (ceil)
        or closest element smaller than value (floor).
    n: int, opt
        Number of elements to be returned, sorted by proximity to the values.
        Default: only the closest value is returned.
    
    Returns
    -------
    Either:
        (output='index'): index/indices of the closest n value(s) in the array;
        (output='value'): the closest n value(s) in the array, 
        (output='both'): closest value(s) and index/-ices, respectively.
    By default, only returns the index/indices.
    
    Possible constraints: 'ceil', 'floor', None ("ceil" will return the closest 
    element with a value greater than 'value', "floor" the opposite)
    """
    if isinstance(array, np.ndarray):
        pass
    elif isinstance(array, list):
        array = np.array(array)
    else:
        raise ValueError("Input type for array should be np.ndarray or list.")
    
    if constraint is None:
        fm = np.absolute(array-value)
        idx = fm.argsort()[:n]
    elif constraint == 'floor' or constraint == 'ceil': 
        indices = np.arange(len(array),dtype=np.int32)
        if constraint == 'floor':
            fm = -(array-value)
        else:
            fm = array-value
        crop_indices = indices[np.where(fm>0)]
        fm = fm[np.where(fm>0)]
        idx = fm.argsort()[:n]
        idx = crop_indices[idx]
        if len(idx)==0:
            msg = "No indices match the constraint ({} w.r.t {:.2f})"
            print(msg.format(constraint,value))
            raise ValueError("No indices match the constraint")
    else:
        raise ValueError("Constraint not recognised")

    if n == 1:
        idx = idx[0]

    if output=='index': return idx
    elif output=='value': return array[idx]
    else: return array[idx], idx


def inject_em_line(wl, flux, lbda, spec, width=None, height=0.1, em=True):
    """
    Injects an emission (or absorption) line in a spectrum. 

    
    Parameters
    ----------
    wl: float
        Wavelength of the line
    flux: float
        Flux of the line to be injected
    lbda : 1d np.ndarray
        Array with the wavelengths (um) of the input spectrum.
    spec : 1d np.ndarray
        Input spectrum fluxes
    width : float, opt
        Full width of the line in mu (see also height). The line will be 
        injected assuming a gaussian profile. If not provided, the width will 
        be set to the 'equivalent width' of the line.
    height: float, opt
        Ratio to peak where the line width is considered. E.g. if 
        height=10%, the width will be the full width at 10% maximum.        
    em: bool, opt
        Whether emission (True) or absorption (False) line.
        
    Returns
    -------
    spec: 1d np.ndarray
        Spectrum with the injected line
    
    """
    # convert ew, assuming it's in mu
    idx_mid = find_nearest(lbda, wl)
    
    nch = len(lbda)
    dlbda = (lbda[idx_mid+1]-lbda[idx_mid-1])/2
    
    # estimate model continuum level using adjacent channels in the spectrum
    lbda_b0 = 0.99*lbda[idx_mid]
    idx_b0 = find_nearest(lbda, lbda_b0, constraint='floor')-1
    lbda_b1 = 0.995*lbda[idx_mid]
    idx_b1 = find_nearest(lbda, lbda_b1, constraint='floor')
    lbda_r1 = 1.01*lbda[idx_mid] 
    idx_r1 = find_nearest(lbda, lbda_r1, constraint='ceil')+1
    lbda_r0 = 1.005*lbda[idx_mid] 
    idx_r0 = find_nearest(lbda, lbda_r0, constraint='ceil')
    if idx_b0<0 or idx_r1>nch-1:
        raise ValueError("The line is too close from the edge of the spectrum")
    spec_b = spec[idx_b0:idx_b1+1]
    spec_r = spec[idx_r0:idx_r1+1]
    cont = np.median(np.concatenate((spec_b,spec_r)))
    
    if width is None:
        # infer ew
        ew = flux/cont
        # infer gaussian profile assuming FWHM = EW
        stddev = ew/(2*np.sqrt(2*np.log(1/height)))
    else:
        stddev = width/(2*np.sqrt(2*np.log(1/height)))
        
    win_sz = int((5*stddev)/dlbda)

    if win_sz%2==0:
        win_sz+=1
    idx_ini = int(idx_mid - (win_sz-1)/2)
    if idx_ini<0:
        msg = "idx ini for line injection negative: try smaller line flux"
        msg+= " than: {} W/m2 (surface flux)"
        raise ValueError(msg.format(flux))
    elif idx_ini+win_sz>len(spec):
        msg = "idx fin for line injection larger than spec length: try smaller"
        msg += " line flux than: {} W/m2 (surface flux)"
        raise ValueError(msg.format(flux))
    if win_sz<1:
        raise ValueError("window size for line injection = {}<1".format(win_sz))
    elif win_sz==1:
        gaus = flux/dlbda
    else:
        gaus = gaussian(win_sz,stddev/dlbda)
        # scale the gaussian to contain exactly required flux
        dlbda_tmp = lbda[idx_ini+1:idx_ini+win_sz+1]-lbda[idx_ini:idx_ini+win_sz]
        gaus = flux*gaus/(np.sum(gaus)*dlbda_tmp)

    
    if em:
        spec[idx_ini:idx_ini+win_sz] += gaus
    else:
        spec[idx_ini:idx_ini+win_sz] -= gaus
    
    return spec

    
def mj_from_rj_and_logg(rp, logg):
    """
    Estimates a planet mass in Jupiter mass for a given radius in Jupiter 
    radius and the log of the surface gravity. 
    
    Parameters
    ----------
    rp: float
        Planet radius in Jupiter radii
    logg: float
        Log of the surface gravity
        
    Returns
    -------
    mj: float
        Planet mass in Jupiter masses
    
    """    
    surf_g = 1e-2 * np.power(10.,logg)  # (m s-2)

    rpJ = rp*c.R_jup.value # (m)

    mp = surf_g*np.power(rpJ,2)/c.G.value # (kg)
    mp /= c.M_jup.value  # (Mjup)

    return mp


def nrefrac(wavelength, density=1.0):
   """
   Calculates refractive index of air from Cauchy formula.
   For comparisong to measurements from the ground, the wavelenghts of model 
   spectra must be slightly shifted using:
   lbda_shift = lbda_model/(1+(nrefrac*1e-6))

   Input: wavelength in Angstrom, 
   Returns N = (n-1) * 1.e6. 
   Credit: France Allard.
   
   Parameters
   ----------
   wavelength: numpy array
       1d numpy array corresponding to the wavelengths of the input spectrum
       in Angstrom
   density: float
       density of air in amagat (relative to STP, e.g. ~10% decrease per 1000m 
       above sea level).
        
   Returns
   -------
   N: float
       Refractive index
   
   """

   # The IAU standard for conversion from air to vacuum wavelengths is given
   # in Morton (1991, ApJS, 77, 119). For vacuum wavelengths (VAC) in
   # Angstroms, convert to air wavelength (AIR) via: 

   #  AIR = VAC / (1.0 + 2.735182E-4 + 131.4182 / VAC^2 + 2.76249E8 / VAC^4)

   try:
       wl = np.array(wavelength)
   except TypeError:
       return None

   wl2inv = (1e4/wl)**2
   refracstp = 272.643 + 1.2288 * wl2inv  + 3.555e-2 * wl2inv**2
   return density * refracstp